import pytest
from dbt.tests.util import run_dbt


snapshots__snapshot_query_tag_sql = """
{% snapshot snapshot_query_tag %}
    {{
        config(
            target_database=database,
            target_schema=schema,
            unique_key='id',
            strategy='check',
            check_cols=['color'],
        )
    }}
    select 1 as id, 'blue' as color
{% endsnapshot %}
"""


models__table_model_query_tag_sql = """
{{ config(materialized = 'table') }}
select 1 as id
"""


models__models_config_yml = """
version: 2

models:
  - name: view_model_query_tag
    columns:
      - name: id
        data_tests:
          - unique
"""


models__view_model_query_tag_sql = """
{{ config(materialized = 'view') }}
select 1 as id
"""


models__incremental_model_query_tag_sql = """
{{ config(materialized = 'incremental', unique_key = 'id') }}
select 1 as id
"""


macros__check_tag_sql = """
{% macro check_query_tag() %}

  {% if execute %}
    {% set query_tag = get_current_query_tag() %}
    {% if query_tag != var("query_tag") %}
      {{ exceptions.raise_compiler_error("Query tag not used!") }}
    {% endif %}
  {% endif %}

{% endmacro %}
"""


seeds__seed_query_tag_csv = """id
1
""".strip()


class TestQueryTag:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "table_model_query_tag.sql": models__table_model_query_tag_sql,
            "view_model_query_tag.sql": models__view_model_query_tag_sql,
            "incremental_model_query_tag.sql": models__incremental_model_query_tag_sql,
            "models_config.yml": models__models_config_yml,
        }

    @pytest.fixture(scope="class")
    def snapshots(self):
        return {"snapshot_query_tag.sql": snapshots__snapshot_query_tag_sql}

    @pytest.fixture(scope="class")
    def macros(self):
        return {"check_tag.sql": macros__check_tag_sql}

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"seed_query_tag.csv": seeds__seed_query_tag_csv}

    @pytest.fixture(scope="class")
    def project_config_update(self, prefix):
        return {
            "config-version": 2,
            "models": {"query_tag": prefix, "post-hook": "{{ check_query_tag() }}"},
            "seeds": {"query_tag": prefix, "post-hook": "{{ check_query_tag() }}"},
            "snapshots": {"query_tag": prefix, "post-hook": "{{ check_query_tag() }}"},
            "tests": {"test": {"query_tag": prefix, "post-hook": "{{ check_query_tag() }}"}},
        }

    def build_all_with_query_tags(self, project, prefix):
        run_dbt(["build", "--vars", '{{"query_tag": "{}"}}'.format(prefix)])

    def test_snowflake_query_tag(self, project, prefix):
        self.build_all_with_query_tags(project, prefix)
        self.build_all_with_query_tags(project, prefix)


class TestSnowflakeProfileQueryTag:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "table_model_query_tag.sql": models__table_model_query_tag_sql,
            "view_model_query_tag.sql": models__view_model_query_tag_sql,
            "incremental_model_query_tag.sql": models__incremental_model_query_tag_sql,
            "models_config.yml": models__models_config_yml,
        }

    @pytest.fixture(scope="class")
    def profiles_config_update(self, prefix):
        return {"query_tag": prefix}

    def build_all_with_query_tags(self, project, prefix):
        run_dbt(["build", "--vars", '{{"query_tag": "{}"}}'.format(prefix)])

    def test_snowflake_query_tag(self, project, prefix):
        self.build_all_with_query_tags(project, prefix)
        self.build_all_with_query_tags(project, prefix)



"""Tests for preserving collation on VARCHAR columns during incremental runs."""

import pytest
from dbt.tests.util import run_dbt, check_relations_equal


# Initial table with collation
_MODEL_BASE_TABLE = """
{{
    config(
        materialized='table',
    )
}}

select 1 as id, COLLATE('initial', 'en-ci')  as some_string_col
"""

# Incremental model that will trigger schema changes
_MODEL_INCREMENTAL_COLLATION = """
{{
    config(
        materialized='incremental',
        unique_key='id',
        on_schema_change='sync_all_columns'
    )
}}

select
    id,
    some_string_col,
    {% if is_incremental() %}
        'new_value' as description
    {% endif %}
from {{ ref('base_table') }}
"""

# Expected result table
_MODEL_EXPECTED = """
{{
    config(materialized='table')
}}

select
    1 as id,
    'initial' as some_string_col,
    'new_value' as description
"""


def _get_column_type(project, table_name, column_name):
    sql = f"""
    describe table {project.database}.{project.test_schema}.{table_name}
    """
    results = project.run_sql(sql, fetch="all")
    for row in results:
        if row[0].lower() == column_name:
            return row[1]
    raise ValueError(f"Column {column_name} not found in table {table_name}")


class TestIncrementalCollation:
    """Test that collation is preserved during incremental runs with schema changes.
    For more info on collation: https://docs.snowflake.com/en/sql-reference/collation
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "base_table.sql": _MODEL_BASE_TABLE,
            "incremental_collation.sql": _MODEL_INCREMENTAL_COLLATION,
            "expected.sql": _MODEL_EXPECTED,
        }

    def test_collation_preserved_on_incremental(self, project):
        # First run: create seed table
        results = run_dbt(["run", "--select", "base_table"])
        assert len(results) == 1

        # Find the 'some_string_col' column and check its type includes collation
        col_type = _get_column_type(project, "base_table", "some_string_col")

        assert col_type is not None, "some_string_col column not found in base_table"
        # Verify collation is present in the type
        assert "COLLATE" in col_type.upper(), f"Expected collation in type but got: {col_type}"

        # First incremental run: create table without description column
        results = run_dbt(["run", "--select", "incremental_collation"])
        assert len(results) == 1

        project.run_sql(
            f"alter table {project.database}.{project.test_schema}.incremental_collation alter column some_string_col set data type varchar(134217728) collate 'en-ci'"
        )

        results = run_dbt(["run", "--select", "incremental_collation"])
        assert len(results) == 1

        # Verify the collation is still present after the incremental run
        col_type_after = _get_column_type(project, "incremental_collation", "some_string_col")
        # Check the 'name' column still has collation
        assert (
            col_type_after is not None
        ), "some_string_col column not found in incremental_collation"
        assert (
            "COLLATE" in col_type_after.upper()
        ), f"Collation was lost after incremental run. Got: {col_type_after}"

        # Verify the data is correct
        run_dbt(["run", "--select", "expected"])
        check_relations_equal(project.adapter, ["incremental_collation", "expected"])


_MODEL_HAS_COLLATION_STG_NO_COLLATION = """
{{
    config(
        materialized='incremental',
        unique_key='id',
        on_schema_change='sync_all_columns')
}}

select * from {{ source('test', 'stg_no_collation') }}
"""

_MODEL_HAS_COLLATION = """
{{
    config(
        materialized='incremental',
        unique_key='id',
        on_schema_change='sync_all_columns')
}}

select * from {{ source('test', 'stg_has_collation') }}
"""

_MODEL_NO_COLLATION = """
{{
    config(
        materialized='incremental',
        unique_key='id',
        on_schema_change='sync_all_columns')
}}

select * from {{ source('test', 'stg_has_collation') }}
"""


class TestIncrementalCollationPreservedOnSchemaChange:
    """Test that collation is preserved during incremental runs with schema changes.
    Tests the following scenarios:
    - Incremental run with schema change on a column with no collation to a column with collation
    - Incremental run with schema change on a column with different collation than source
    - Incremental run with schema change on a column with collation to a column with no collation
    To do this we simulate a schema change by creating the tables ahead of time and specifying the collation on the source / target tables.
    Then when we run the models dbt will detect the schema change and apply the correct collation to the target table.
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model_stg_no_collation.sql": _MODEL_HAS_COLLATION_STG_NO_COLLATION,
            "has_collation.sql": _MODEL_HAS_COLLATION,
            "no_collation.sql": _MODEL_NO_COLLATION,
            "schema.yml": """
                            sources:
                            - name: test
                              schema: "{{ target.schema }}"
                              tables:
                                - name: stg_no_collation
                                - name: stg_has_collation
                            """,
        }

    def test_collation_preserved_on_schema_change(self, project):
        # create all the tables ahead of time so we can specify the collation on the source / target tables
        project.run_sql(
            f"""
        create or replace TABLE {project.database}.{project.test_schema}.model_stg_no_collation (
            ID VARCHAR(5),
            ORDER_DATE DATE,
            STATUS VARCHAR(8) COLLATE 'en-ci',
            ORDER_ID VARCHAR(5)
        );"""
        )

        project.run_sql(
            f"""
        create or replace TABLE {project.database}.{project.test_schema}.STG_NO_COLLATION (
            ID VARCHAR(5),
            ORDER_DATE DATE,
            STATUS VARCHAR(10),
            ORDER_ID VARCHAR(5)
        );"""
        )
        project.run_sql(
            f"""
        create or replace TABLE {project.database}.{project.test_schema}.STG_HAS_COLLATION (
            ID VARCHAR(5),
            NAME VARCHAR(12) COLLATE 'en-ci',
            AGE INT
        );"""
        )
        project.run_sql(
            f"""
        create or replace TABLE {project.database}.{project.test_schema}.HAS_COLLATION (
            ID VARCHAR(5),
            NAME VARCHAR(10) COLLATE 'en-cs',
            AGE INT
        );"""
        )
        project.run_sql(
            f"""
        create or replace TABLE {project.database}.{project.test_schema}.NO_COLLATION (
            ID VARCHAR(5),
            NAME VARCHAR(10),
            AGE INT
        );"""
        )

        results = run_dbt(["run"])
        results = run_dbt(["run"])
        assert len(results) == 3

        # Find the various test columns and check the collation is still present
        col_type_status = _get_column_type(project, "model_stg_no_collation", "status")

        assert (
            "COLLATE" in col_type_status.upper()
        ), f"Collation was lost after incremental run. Got: {col_type_status}"

        assert (
            "en-ci" in col_type_status
        ), f"Collation was not set to en-ci. Got: {col_type_status}"

        col_type_name = _get_column_type(project, "has_collation", "name")

        assert col_type_name.lower() == "varchar(12) collate 'en-cs'"

        col_type_name_no_collation = _get_column_type(project, "no_collation", "name")

        assert col_type_name_no_collation.lower() == "varchar(12)"



from dataclasses import dataclass

from dbt.adapters.base.relation import Policy
from dbt_common.dataclass_schema import StrEnum


class SnowflakeRelationType(StrEnum):
    Table = "table"
    View = "view"
    CTE = "cte"
    External = "external"
    DynamicTable = "dynamic_table"
    Function = "function"


class SnowflakeIncludePolicy(Policy):
    database: bool = True
    schema: bool = True
    identifier: bool = True


@dataclass
class SnowflakeQuotePolicy(Policy):
    database: bool = False
    schema: bool = False
    identifier: bool = False



import pytest
from dbt.tests.util import run_dbt


models__test_model_sql = """
{{ config(materialized = 'table') }}
select 1 as id
"""


class TestS3StageVpceDnsName:
    """
    Test that the s3_stage_vpce_dns_name session parameter is properly set
    when specified in the profile configuration.
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_model.sql": models__test_model_sql,
        }

    @pytest.fixture(scope="class")
    def profiles_config_update(self, prefix, dbt_profile_target):
        outputs = {"default": dbt_profile_target}
        outputs["default"][
            "s3_stage_vpce_dns_name"
        ] = f"vpce-{prefix}-xxxxxxxx.s3.us-east-1.vpce.amazonaws.com"

    def test_s3_stage_vpce_dns_name_set(self, project, prefix):
        """Test that the s3_stage_vpce_dns_name session parameter is properly set."""
        # Run dbt to create the model
        run_dbt(["run"])

        # Check that the session parameter was set correctly
        result = project.run_sql(
            "show parameters like 'S3_STAGE_VPCE_DNS_NAME' in session", fetch="one"
        )
        assert result is not None
        expected_value = f"vpce-{prefix}-xxxxxxxx.s3.us-east-1.vpce.amazonaws.com"
        assert result[1] == expected_value, f"Expected {expected_value}, got {result[1]}"



from dbt.tests.adapter.empty.test_empty import (
    BaseTestEmpty,
    BaseTestEmptyInlineSourceRef,
    MetadataWithEmptyFlag,
)


class TestSnowflakeEmpty(BaseTestEmpty):
    pass


class TestSnowflakeEmptyInlineSourceRef(BaseTestEmptyInlineSourceRef):
    pass


class TestMetadataWithEmptyFlag(MetadataWithEmptyFlag):
    pass



from dataclasses import dataclass
import re
from typing import Optional

from dbt.adapters.base.column import Column
from dbt_common.exceptions import DbtRuntimeError

COLLATE_PATTERN = re.compile(r"collate\s+'([^']+)'(\s*rtrim)?", re.IGNORECASE)


@dataclass
class SnowflakeColumn(Column):
    collation: Optional[str] = None

    def is_integer(self) -> bool:
        # everything that smells like an int is actually a NUMBER(38, 0)
        return False

    def is_numeric(self) -> bool:
        return self.dtype.lower() in [
            "int",
            "integer",
            "bigint",
            "smallint",
            "tinyint",
            "byteint",
            "numeric",
            "decimal",
            "number",
        ]

    def is_float(self):
        return self.dtype.lower() in [
            "float",
            "float4",
            "float8",
            "double",
            "double precision",
            "real",
        ]

    def string_size(self) -> int:
        if not self.is_string():
            raise DbtRuntimeError("Called string_size() on non-string field!")

        if self.dtype == "text" or self.char_size is None:
            return 16777216
        else:
            return int(self.char_size)

    @property
    def expanded_data_type(self) -> str:
        """Include collation clause for string types where present."""
        if self.collation and self.is_string():
            return f"{self.data_type} collate '{self.collation}'"

        return self.data_type

    @classmethod
    def from_description(cls, name: str, raw_data_type: str) -> "SnowflakeColumn":
        """
        Parse column information from raw data type string.
        Handles Snowflake-specific syntax including COLLATE clause.

        Examples:
            VARCHAR(16777216) COLLATE 'en-ci-rtrim'
            VARCHAR(100)
            NUMBER(38,0)
        """

        # We want to pass through numeric parsing for composite types
        if raw_data_type.lower().startswith(("array", "object", "map", "vector")):
            return cls(name, raw_data_type, None, None, None)

        collation = None
        # Check if there's a COLLATE clause and extract it
        if raw_data_type.lower().startswith(
            ("varchar", "character varying", "character", "varchar", "text")
        ):
            collate_match = COLLATE_PATTERN.search(raw_data_type, re.IGNORECASE)
            collation = collate_match.group(1) if collate_match else None

        # Parse the base type using parent class logic
        raw_data_type_without_collation = COLLATE_PATTERN.sub("", raw_data_type).strip()
        base_column = super().from_description(name, raw_data_type_without_collation)

        # Create a SnowflakeColumn with the parsed information plus collation
        return cls(
            column=base_column.column,
            dtype=base_column.dtype,
            char_size=base_column.char_size,
            numeric_precision=base_column.numeric_precision,
            numeric_scale=base_column.numeric_scale,
            collation=collation,
        )

    def is_array(self) -> bool:
        return self.dtype.lower().startswith("array")

    def is_object(self) -> bool:
        return self.dtype.lower().startswith("object")

    def is_map(self) -> bool:
        return self.dtype.lower().startswith("map")



from typing import Iterable, List, Optional, Union

from dbt_common.exceptions import DbtConfigError

from dbt.adapters.catalogs import CATALOG_INTEGRATION_MODEL_CONFIG_NAME
from dbt.adapters.contracts.relation import RelationConfig

from dbt.adapters.snowflake import constants
from dbt.adapters.snowflake.constants import SnowflakeIcebergTableRelationParameters


def auto_refresh(model: RelationConfig) -> Optional[bool]:
    return model.config.get("auto_refresh") if model.config else None


def max_data_extension_time_in_days(model: RelationConfig) -> Optional[int]:
    return (
        model.config.get(
            SnowflakeIcebergTableRelationParameters.max_data_extension_time_in_days, False
        )
        if model.config
        else None
    )


def automatic_clustering(model: RelationConfig) -> Optional[bool]:
    return (
        model.config.get(SnowflakeIcebergTableRelationParameters.automatic_clustering, False)
        if model.config
        else None
    )


def base_location(model: RelationConfig) -> Optional[str]:
    if not model.config:
        return None

    prefix = (
        model.config.get("base_location_root") or "_dbt"
    )  # use "_dbt" even when users pass in None
    path = f"{prefix}/{model.schema}/{model.identifier}"
    if subpath := model.config.get("base_location_subpath"):
        path += f"/{subpath}"

    return path


def catalog_name(model: RelationConfig) -> Optional[str]:
    if not model.config or not hasattr(model.config, "get"):
        return None

    if _catalog := model.config.get(CATALOG_INTEGRATION_MODEL_CONFIG_NAME):
        # make catalog_name case-insensitive
        return _catalog.upper()

    _table_format = table_format(model)
    if _table_format == constants.ICEBERG_TABLE_FORMAT:
        return constants.DEFAULT_BUILT_IN_CATALOG.name

    return constants.DEFAULT_INFO_SCHEMA_CATALOG.name


def cluster_by(model: RelationConfig) -> Optional[str]:
    if not model.config:
        return None

    fields = model.config.get("cluster_by")
    if isinstance(fields, str):
        return fields
    if isinstance(fields, Iterable):
        return ", ".join(fields)
    if fields is not None:
        raise DbtConfigError(f"Unexpected cluster_by configuration: {fields}")
    return None


# Keys may have to be lowercased due to Glue
def partition_by(model: RelationConfig) -> Optional[Union[str, List[str]]]:
    if not model.config:
        return None

    fields = model.config.get("partition_by")
    if isinstance(fields, str):
        return fields
    if isinstance(fields, Iterable):
        return list(fields)
    if fields is not None:
        raise DbtConfigError(f"Unexpected partition_by configuration: {fields}")
    return None


def external_volume(model: RelationConfig) -> Optional[str]:
    return model.config.get("external_volume") if model.config else None


def is_transient(model: RelationConfig) -> Optional[bool]:
    """
    Always supply transient on table create DDL unless user specifically sets
    transient to false or unset.

    Args:
        model (RelationConfig): `config.model` (not `model`) from the jinja context.

    Returns:
        None if there is no materialized config on the model; this shouldn't happen
        False if we know this is an iceberg table format (excludes format set by the catalog)
        True if the user has set it to True or if the user has explicitly unset it
        False otherwise
    """
    if not model.config:
        return None

    if table_format(model) == constants.ICEBERG_TABLE_FORMAT:
        return False
    return model.config.get("transient", False) or model.config.get("transient", True)


def table_format(model: RelationConfig) -> Optional[str]:
    if not model.config:
        return None

    # we don't know what the table format is if it's not on the model
    # this could be derived from the catalog and will be derived from the catalog moving forward
    # so we cannot default to INFO_SCHEMA here
    if _table_format := model.config.get("table_format"):
        # make table_format case-insensitive
        return _table_format.upper()
    return None


def target_file_size(model: RelationConfig) -> Optional[str]:
    return model.config.get("target_file_size") if model.config else None



_MODEL_BASIC_TABLE_MODEL = """
{{
  config(
    materialized = "table",
    cluster_by=['id'],
  )
}}
select 1 as id
"""

_MODEL_BASIC_TABLE_LITERALS = """
{{
  config(
    materialized = "table",
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
  )
}}

select
    object_construct('theme', 'dark', 'notifications', TO_VARCHAR(true))::MAP(VARCHAR(20), VARCHAR(20)) as my_map_conversion,
"""

_MODEL_BASIC_ICEBERG_MODEL = """
{{
  config(
    transient = "true",
    materialized = "table",
    cluster_by=['id'],
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
  )
}}

select * from {{ ref('first_table') }}
"""

_MODEL_BASIC_ICEBERG_BUILTIN_MODEL = """
{{
  config(
    transient = "true",
    materialized = "table",
    cluster_by=['id'],
    catalog="snowflake",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
  )
}}

select * from {{ ref('first_table') }}
"""

_MODEL_BASIC_ICEBERG_MODEL_WITH_PATH = """
{{
  config(
    transient = "true",
    materialized = "table",
    cluster_by=['id'],
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_root="root_path",
  )
}}

select * from {{ ref('first_table') }}
"""

_MODEL_BASIC_ICEBERG_MODEL_WITH_PATH_SUBPATH = """
{{
  config(
    transient = "true",
    materialized = "table",
    cluster_by=['id'],
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_root="root_path",
    base_location_subpath="subpath",
  )
}}

select * from {{ ref('first_table') }}
"""

_MODEL_BASIC_DYNAMIC_TABLE_MODEL = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='1 minute',
    refresh_mode='INCREMENTAL',
    table_format='iceberg',
    external_volume='s3_iceberg_snow',
) }}

select * from {{ ref('first_table') }}
"""

_MODEL_BASIC_DYNAMIC_TABLE_ICEBERG_MODEL = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='1 minute',
    refresh_mode='INCREMENTAL',
    catalog='snowflake',
    external_volume='s3_iceberg_snow',
) }}

select * from {{ ref('first_table') }}
"""

_MODEL_BASIC_DYNAMIC_TABLE_MODEL_WITH_PATH = """
{{
  config(
    materialized = "dynamic_table",
    snowflake_warehouse='DBT_TESTING',
    target_lag='1 minute',
    cluster_by=['id'],
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_root="root_path",
  )
}}

select * from {{ ref('first_table') }}
"""

_MODEL_BASIC_DYNAMIC_TABLE_MODEL_WITH_PATH_SUBPATH = """
{{
  config(
    materialized = "dynamic_table",
    snowflake_warehouse='DBT_TESTING',
    target_lag='1 minute',
    cluster_by=['id'],
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_root="root_path",
    base_location_subpath='subpath',
  )
}}

select * from {{ ref('first_table') }}
"""


_MODEL_BASIC_DYNAMIC_TABLE_MODEL_WITH_SUBPATH = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='1 minute',
    refresh_mode='INCREMENTAL',
    table_format='iceberg',
    external_volume='s3_iceberg_snow',
    base_location_subpath='subpath',
) }}

select * from {{ ref('first_table') }}
"""

_MODEL_BUILT_ON_ICEBERG_TABLE = """
{{
  config(
    materialized = "table",
  )
}}
select * from {{ ref('iceberg_table') }}
"""

_MODEL_TABLE_BEFORE_SWAP = """
{{
  config(
    materialized = "table",
  )
}}
select 1 as id
"""

_MODEL_VIEW_BEFORE_SWAP = """
select 1 as id
"""

_MODEL_TABLE_FOR_SWAP_ICEBERG = """
{{
  config(
    materialized = "table",
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
  )
}}
select 1 as id
"""



import pytest
from dbt.tests.adapter.incremental.test_incremental_microbatch import (
    BaseMicrobatch,
)
from dbt.tests.util import run_dbt

# Create input with UTC timestamps
_input_model_sql = """
{{ config(materialized='table', event_time='event_time') }}
select 1 as id, to_timestamp_tz('2020-01-01 00:00:00-0') as event_time
union all
select 2 as id, to_timestamp_tz('2020-01-02 00:00:00-0') as event_time
union all
select 3 as id, to_timestamp_tz('2020-01-03 00:00:00-0') as event_time
"""

# Create input with +1 tz timestamp
_input_model_with_offset_sql = """
{{ config(materialized='table', event_time='event_time') }}
select 3 as id, '2020-01-03 00:37:00+01:00' as event_time
"""


# No requirement for a unique_id for snowflake microbatch!
_microbatch_model_no_unique_id_sql = """
{{ config(materialized='incremental', incremental_strategy='microbatch', event_time='event_time', batch_size='day', begin=modules.datetime.datetime(2020, 1, 1, 0, 0, 0)) }}
select * from {{ ref('input_model') }}
"""

_microbatch_model_hourly_sql = """
{{ config(materialized='incremental', incremental_strategy='microbatch', event_time='event_time', batch_size='hour', begin=modules.datetime.datetime(2020, 1, 1, 0, 0, 0)) }}
select * from {{ ref('input_model') }}
"""


class TestSnowflakeMicrobatch(BaseMicrobatch):
    @pytest.fixture(scope="class")
    def microbatch_model_sql(self) -> str:
        return _microbatch_model_no_unique_id_sql

    @pytest.fixture(scope="class")
    def input_model_sql(self) -> str:
        return _input_model_sql

    @pytest.fixture(scope="class")
    def insert_two_rows_sql(self, project) -> str:
        test_schema_relation = project.adapter.Relation.create(
            database=project.database, schema=project.test_schema
        )
        return f"insert into {test_schema_relation}.input_model (id, event_time) values (4, '2020-01-04 00:00:00-0'), (5, '2020-01-05 00:00:00-0')"


class TestSnowflakeMicrobatchConvertsStartEndTimeToTimestamp(TestSnowflakeMicrobatch):
    @pytest.fixture(scope="class")
    def input_model_sql(self) -> str:
        return _input_model_with_offset_sql

    @pytest.fixture(scope="class")
    def microbatch_model_sql(self) -> str:
        return _microbatch_model_hourly_sql

    def test_run_with_event_time(self, project):
        run_dbt(
            [
                "run",
                "--debug",
                "--event-time-start",
                "2020-01-03 00:00:00",
                "--event-time-end",
                "2020-01-03 01:00:00",
            ]
        )
        self.assert_row_count(project, "input_model", 1)
        self.assert_row_count(project, "microbatch_model", 0)



from dataclasses import dataclass
from typing import Optional, Union

from dbt.adapters.catalogs import (
    CatalogIntegration,
    CatalogIntegrationConfig,
)
from dbt.adapters.catalogs import InvalidCatalogIntegrationConfigError
from dbt.adapters.contracts.relation import RelationConfig
from dbt.adapters.snowflake import constants, parse_model

from dbt.adapters.exceptions.compilation import InvalidRelationConfigError


@dataclass
class IcebergRestCatalogRelation:
    catalog_type: str = constants.DEFAULT_ICEBERG_REST_CATALOG.catalog_type
    catalog_name: Optional[str] = constants.DEFAULT_ICEBERG_REST_CATALOG.name
    table_format: Optional[str] = constants.ICEBERG_TABLE_FORMAT
    partition_by: Optional[Union[list[str], str]] = None
    catalog_linked_database: Optional[str] = None
    catalog_linked_database_type: Optional[str] = None  # e.g., 'glue' for AWS Glue
    external_volume: Optional[str] = None
    file_format: Optional[str] = None
    target_file_size: Optional[str] = None
    max_data_extension_time_in_days: Optional[int] = None
    auto_refresh: Optional[bool] = None


class IcebergRestCatalogIntegration(CatalogIntegration):
    catalog_type = constants.DEFAULT_ICEBERG_REST_CATALOG.catalog_type
    catalog_name: Optional[str] = constants.DEFAULT_ICEBERG_REST_CATALOG.name
    table_format = constants.ICEBERG_TABLE_FORMAT
    file_format = None  # Snowflake chooses based on stage-format
    allows_writes = True
    auto_refresh = None
    catalog_linked_database: Optional[str] = None
    catalog_linked_database_type: Optional[str] = None
    max_data_extension_time_in_days: Optional[int] = None
    target_file_size: Optional[str] = None

    def __init__(self, config: CatalogIntegrationConfig) -> None:
        # we overwrite this because the base provides too much config
        self.name: str = config.name
        self.catalog_name: Optional[str] = config.catalog_name
        self.external_volume: Optional[str] = config.external_volume
        if adapter_properties := config.adapter_properties:
            self.catalog_linked_database = adapter_properties.get("catalog_linked_database")
            self.catalog_linked_database_type = adapter_properties.get(
                "catalog_linked_database_type"
            )
            self.auto_refresh = adapter_properties.get("auto_refresh")
            self.target_file_size = adapter_properties.get("target_file_size")
            self.max_data_extension_time_in_days = adapter_properties.get(
                "max_data_extension_time_in_days"
            )

        if not self.catalog_linked_database:
            raise InvalidCatalogIntegrationConfigError(
                config.name,
                "adapter_properties.catalog_linked_database is currently required for iceberg rest catalog integrations with catalog linked databases",
            )

    def build_relation(self, model: RelationConfig) -> IcebergRestCatalogRelation:
        """
        Args:
            model: `config.model` (not `model`) from the jinja context
        """
        cluster_by = parse_model.cluster_by(model)
        if cluster_by:
            raise InvalidRelationConfigError(
                model,
                model.config,
                "cluster_by is not supported for iceberg rest catalog integrations with catalog linked databases",
            )

        if parse_model.external_volume(model):
            raise InvalidRelationConfigError(
                model,
                model.config,
                "external_volume is not supported for iceberg rest catalog integrations with catalog linked databases",
            )

        return IcebergRestCatalogRelation(
            catalog_name=self.name,
            external_volume=None,
            partition_by=parse_model.partition_by(model),
            catalog_linked_database=self.catalog_linked_database,
            catalog_linked_database_type=self.catalog_linked_database_type,
            auto_refresh=parse_model.auto_refresh(model) or self.auto_refresh,
            target_file_size=parse_model.target_file_size(model) or self.target_file_size,
            max_data_extension_time_in_days=parse_model.max_data_extension_time_in_days(model)
            or self.max_data_extension_time_in_days,
        )



SEED = """
id,value
1,100
2,200
3,300
""".strip()


TABLE = """
{{ config(
    materialized='table',
) }}
select * from {{ ref('my_seed') }}
"""


VIEW = """
{{ config(
    materialized='view',
) }}
select * from {{ ref('my_seed') }}
"""


DYNAMIC_TABLE = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='1 minute',
    refresh_mode='INCREMENTAL',
) }}
select * from {{ ref('my_seed') }}
"""


DYNAMIC_ICEBERG_TABLE = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='1 minute',
    refresh_mode='INCREMENTAL',
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
) }}
select * from {{ ref('my_seed') }}
"""

ICEBERG_TABLE = """
{{ config(
    materialized='table',
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
) }}
select * from {{ ref('my_seed') }}
"""

INCREMENTAL_ICEBERG_TABLE = """
{{ config(
    materialized='incremental',
    table_format='iceberg',
    incremental_strategy='append',
    unique_key="id",
    external_volume = "s3_iceberg_snow",
) }}
select * from {{ ref('my_seed') }}
"""


INCREMENTAL_TABLE = """
{{ config(
    materialized='incremental',
    incremental_strategy='append',
    unique_key="id",
) }}
select * from {{ ref('my_seed') }}
"""



from dbt.adapters.snowflake.relation_configs.dynamic_table import (
    RefreshMode,
    SnowflakeDynamicTableConfig,
    SnowflakeDynamicTableConfigChangeset,
    SnowflakeDynamicTableRefreshModeConfigChange,
    SnowflakeDynamicTableWarehouseConfigChange,
    SnowflakeDynamicTableTargetLagConfigChange,
)
from dbt.adapters.snowflake.relation_configs.policies import (
    SnowflakeIncludePolicy,
    SnowflakeQuotePolicy,
    SnowflakeRelationType,
)



import pytest
from dbt.tests.adapter.utils.test_timestamps import BaseCurrentTimestamps

_MODEL_CURRENT_TIMESTAMP = """
SELECT {{current_timestamp()}} as current_timestamp,
       {{current_timestamp_in_utc_backcompat()}} as current_timestamp_in_utc_backcompat,
       {{current_timestamp_backcompat()}} as current_timestamp_backcompat
"""


class TestCurrentTimestampSnowflake(BaseCurrentTimestamps):
    @pytest.fixture(scope="class")
    def models(self):
        return {"get_current_timestamp.sql": _MODEL_CURRENT_TIMESTAMP}

    @pytest.fixture(scope="class")
    def expected_schema(self):
        return {
            "CURRENT_TIMESTAMP": "TIMESTAMP_TZ",
            "CURRENT_TIMESTAMP_IN_UTC_BACKCOMPAT": "TIMESTAMP_NTZ",
            "CURRENT_TIMESTAMP_BACKCOMPAT": "TIMESTAMP_NTZ",
        }

    @pytest.fixture(scope="class")
    def expected_sql(self):
        return """
                select convert_timezone('UTC', current_timestamp()) as current_timestamp,
                       convert_timezone('UTC', current_timestamp::TIMESTAMP)::TIMESTAMP as current_timestamp_in_utc_backcompat,
                       current_timestamp::TIMESTAMP as current_timestamp_backcompat
                """



import pytest

from dbt.tests.adapter.constraints.test_constraints import (
    BaseTableConstraintsColumnsEqual,
    BaseViewConstraintsColumnsEqual,
    BaseTableContractSqlHeader,
    BaseIncrementalContractSqlHeader,
    BaseIncrementalConstraintsColumnsEqual,
    BaseConstraintsRuntimeDdlEnforcement,
    BaseConstraintsRollback,
    BaseIncrementalConstraintsRuntimeDdlEnforcement,
    BaseIncrementalConstraintsRollback,
    BaseModelConstraintsRuntimeEnforcement,
    BaseConstraintQuotedColumn,
)

from dbt.tests.adapter.constraints.fixtures import (
    model_contract_header_schema_yml,
)

my_model_contract_sql_header_sql = """
{{
  config(
    materialized = "table"
  )
}}
{% call set_sql_header(config) %}
SET MY_VARIABLE='test';
{% endcall %}
SELECT $MY_VARIABLE as column_name
"""

my_model_incremental_contract_sql_header_sql = """
{{
  config(
    materialized = "incremental",
    on_schema_change="append_new_columns"
  )
}}
{% call set_sql_header(config) %}
SET MY_VARIABLE='test';
{% endcall %}
SELECT $MY_VARIABLE as column_name
"""

_expected_sql_snowflake = """
create or replace transient table <model_identifier> (
    id integer not null primary key references <foreign_key_model_identifier> (id) unique,
    color text,
    date_day text
) as ( select
        id,
        color,
        date_day from
    (
    -- depends_on: <foreign_key_model_identifier>
    select
        'blue' as color,
        1 as id,
        '2019-01-01' as date_day
    ) as model_subq
);
"""


class SnowflakeColumnEqualSetup:
    @pytest.fixture
    def int_type(self):
        return "FIXED"

    @pytest.fixture
    def schema_int_type(self):
        return "INT"

    @pytest.fixture
    def data_types(self, int_type, schema_int_type, string_type):
        # sql_column_value, schema_data_type, error_data_type
        return [
            ["1", schema_int_type, int_type],
            ["'1'", string_type, string_type],
            ["cast('2019-01-01' as date)", "date", "DATE"],
            ["true", "boolean", "BOOLEAN"],
            ["'2013-11-03 00:00:00-07'::timestamptz", "timestamp_tz", "TIMESTAMP_TZ"],
            ["'2013-11-03 00:00:00-07'::timestamp", "timestamp", "TIMESTAMP_NTZ"],
            ["ARRAY_CONSTRUCT('a','b','c')", "array", "ARRAY"],
            ["ARRAY_CONSTRUCT(1,2,3)", "array", "ARRAY"],
            ["TO_GEOGRAPHY('POINT(-122.35 37.55)')", "geography", "GEOGRAPHY"],
            ["TO_GEOMETRY('POINT(1820.12 890.56)')", "geometry", "GEOMETRY"],
            [
                """TO_VARIANT(PARSE_JSON('{"key3": "value3", "key4": "value4"}'))""",
                "variant",
                "VARIANT",
            ],
        ]


class TestSnowflakeTableConstraintsColumnsEqual(
    SnowflakeColumnEqualSetup, BaseTableConstraintsColumnsEqual
):
    pass


class TestSnowflakeViewConstraintsColumnsEqual(
    SnowflakeColumnEqualSetup, BaseViewConstraintsColumnsEqual
):
    pass


class TestSnowflakeIncrementalConstraintsColumnsEqual(
    SnowflakeColumnEqualSetup, BaseIncrementalConstraintsColumnsEqual
):
    pass


class TestSnowflakeTableContractsSqlHeader(BaseTableContractSqlHeader):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_contract_sql_header.sql": my_model_contract_sql_header_sql,
            "constraints_schema.yml": model_contract_header_schema_yml,
        }


class TestSnowflakeIncrementalContractsSqlHeader(BaseIncrementalContractSqlHeader):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_contract_sql_header.sql": my_model_incremental_contract_sql_header_sql,
            "constraints_schema.yml": model_contract_header_schema_yml,
        }


class TestSnowflakeTableConstraintsDdlEnforcement(BaseConstraintsRuntimeDdlEnforcement):
    @pytest.fixture(scope="class")
    def expected_sql(self):
        return _expected_sql_snowflake


class TestSnowflakeIncrementalConstraintsDdlEnforcement(
    BaseIncrementalConstraintsRuntimeDdlEnforcement
):
    @pytest.fixture(scope="class")
    def expected_sql(self):
        return _expected_sql_snowflake


class TestSnowflakeTableConstraintsRollback(BaseConstraintsRollback):
    @pytest.fixture(scope="class")
    def expected_error_messages(self):
        return ["NULL result in a non-nullable column"]


class TestSnowflakeIncrementalConstraintsRollback(BaseIncrementalConstraintsRollback):
    @pytest.fixture(scope="class")
    def expected_error_messages(self):
        return ["NULL result in a non-nullable column"]


class TestSnowflakeModelConstraintsRuntimeEnforcement(BaseModelConstraintsRuntimeEnforcement):
    @pytest.fixture(scope="class")
    def expected_sql(self):
        return """
create or replace transient table <model_identifier> (
    id integer not null,
    color text,
    date_day text,
    primary key (id),
    constraint strange_uniqueness_requirement unique (color, date_day),
    foreign key (id) references <foreign_key_model_identifier> (id)
) as ( select
        id,
        color,
        date_day from
    (
    -- depends_on: <foreign_key_model_identifier>
    select
        'blue' as color,
        1 as id,
        '2019-01-01' as date_day
    ) as model_subq
);
"""


class TestSnowflakeConstraintQuotedColumn(BaseConstraintQuotedColumn):
    @pytest.fixture(scope="class")
    def expected_sql(self):
        return """
create or replace transient table <model_identifier> (
    id integer not null,
    "from" text not null,
    date_day text
) as (
    select id, "from", date_day
    from (
        select
          'blue' as "from",
          1 as id,
          '2019-01-01' as date_day
    ) as model_subq
);
"""



from dbt.contracts.results import CatalogArtifact
from dbt.tests.util import run_dbt
import pytest

from tests.functional.adapter.catalog_tests import files


class TestCatalogRelationTypes:
    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {"my_seed.csv": files.MY_SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        yield {
            "my_table.sql": files.MY_TABLE,
            "my_view.sql": files.MY_VIEW,
            "my_dynamic_table.sql": files.MY_DYNAMIC_TABLE,
        }

    @pytest.fixture(scope="class", autouse=True)
    def docs(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])
        yield run_dbt(["docs", "generate"])

    @pytest.mark.parametrize(
        "node_name,relation_type",
        [
            ("seed.test.my_seed", "BASE TABLE"),
            ("model.test.my_table", "BASE TABLE"),
            ("model.test.my_view", "VIEW"),
            ("model.test.my_dynamic_table", "DYNAMIC TABLE"),
        ],
    )
    def test_relation_types_populate_correctly(
        self, docs: CatalogArtifact, node_name: str, relation_type: str
    ):
        """
        This test addresses: https://github.com/dbt-labs/dbt-snowflake/issues/817
        """
        assert node_name in docs.nodes
        node = docs.nodes[node_name]
        assert node.metadata.type == relation_type



import pytest

from dbt.tests.util import (
    run_dbt,
)

_MODELS__view = """
{{ config(
    materialized='table',
) }}

with dupes as (
    select 'foo' as key, 1 as value
    union all
    select 'foo' as key, 2 as value
)

select
    object_agg(key, value) as agg
from dupes
"""


class TestDuplicateKeyNotInExceptions:
    @pytest.fixture(scope="class")
    def models(self):
        return {"model.sql": _MODELS__view}

    def test_row_values_were_scrubbed_from_duplicate_merge_exception(self, project):
        result = run_dbt(["run", "-s", "model"], expect_pass=False)
        assert len(result) == 1
        assert "Duplicate field key '[redacted]'" in result[0].message
        assert "'foo'" not in result[0].message



import os

import pytest
import re

from dbt.tests.util import run_dbt


def get_cleanded_model_ddl_from_file(file_name: str) -> str:
    with open(f"target/run/test/models/{file_name}", "r") as ddl_file:
        return re.sub(r"\s+", " ", ddl_file.read())


_MY_SEED = """
id,value
1,100
2,200
3,300
""".strip()

_CREATE_TABLE_ROW_ACCESS_POLICY = """
create or replace row access policy always_true as (id integer) returns boolean ->
  case
      when id = 1 then true
      else false
  end
;
"""

_CREATE_TABLE_TAG = """
create or replace tag tag_name COMMENT = 'testing'
"""

_MODELS_TABLE_ROW_ACCESS_POLICY = """
{{ config(
    materialized = 'table',
    row_access_policy = 'always_true on (id)',
) }}

select 1 as id
"""
_DDL_MODELS_TABLE_ROW_ACCESS_POLICY = (
    """.table_row_access_policy with row access policy always_true on (id) as"""
)

_MODELS_TABLE_TAG = """
{{ config(
    materialized = 'table',
    table_tag = "tag_name = 'tag_value'",
) }}

select 1 as id
"""
_DDL_MODELS_TABLE_TAG = """.table_tag with tag (tag_name = 'tag_value')"""

_MODELS_TABLE_TAG_AND_ROW_ACCESS_POLICY = """
{{ config(
    materialized = 'table',
    table_tag = "tag_name = 'tag_value'",
    row_access_policy = 'always_true on (id)',
) }}

select 1 as id
"""
_DDL_MODELS_TABLE_TAG_AND_ROW_ACCESS_POLICY = """.table_tag_and_row_access_policy with row access policy always_true on (id) with tag (tag_name = 'tag_value') as"""

# View model templates
_MODELS_VIEW_ROW_ACCESS_POLICY = """
{{ config(
    materialized = 'view',
    row_access_policy = 'always_true on (id)',
) }}

select 1 as id
"""
_DDL_MODELS_VIEW_ROW_ACCESS_POLICY = (
    """.view_row_access_policy with row access policy always_true on (id) as"""
)

_MODELS_VIEW_TAG = """
{{ config(
    materialized = 'view',
    table_tag = "tag_name = 'tag_value'",
) }}

select 1 as id
"""
_DDL_MODELS_VIEW_TAG = """view_tag with tag (tag_name = 'tag_value')"""

_MODELS_VIEW_TAG_AND_ROW_ACCESS_POLICY = """
{{ config(
    materialized = 'view',
    table_tag = "tag_name = 'tag_value'",
    row_access_policy = 'always_true on (id)',
) }}

select 1 as id
"""
_DDL_MODELS_VIEW_TAG_AND_ROW_ACCESS_POLICY = """view_tag_and_row_access_policy with row access policy always_true on (id) with tag (tag_name = 'tag_value') as"""

# Incremental model templates
_MODELS_INCREMENTAL_ROW_ACCESS_POLICY = """
{{ config(
    materialized = 'incremental',
    row_access_policy = 'always_true on (id)',
) }}

select 1 as id
"""
_DDL_MODELS_INCREMENTAL_ROW_ACCESS_POLICY = (
    """.incremental_row_access_policy with row access policy always_true on (id) as"""
)

_MODELS_INCREMENTAL_TAG = """
{{ config(
    materialized = 'incremental',
    table_tag = "tag_name = 'tag_value'",
) }}

select 1 as id
"""
_DDL_MODELS_INCREMENTAL_TAG = """.incremental_tag with tag (tag_name = 'tag_value')"""

_MODELS_INCREMENTAL_TAG_AND_ROW_ACCESS_POLICY = """
{{ config(
    materialized = 'incremental',
    table_tag = "tag_name = 'tag_value'",
    row_access_policy = 'always_true on (id)',
) }}

select 1 as id
"""
_DDL_MODELS_INCREMENTAL_TAG_AND_ROW_ACCESS_POLICY = """.incremental_tag_and_row_access_policy with row access policy always_true on (id) with tag (tag_name = 'tag_value') as"""

# Dynamic table model templates
_MODELS_DYNAMIC_TABLE_ROW_ACCESS_POLICY = (
    """
{{ config(
    target_lag = '7 days',
    snowflake_warehouse = '"""
    + os.getenv("SNOWFLAKE_TEST_WAREHOUSE", "")
    + """',
    materialized = 'dynamic_table',
    row_access_policy = 'always_true on (id)',
) }}

select * from {{ ref("my_seed") }}
"""
)
_DDL_MODELS_DYNAMIC_TABLE_ROW_ACCESS_POLICY = (
    r"\.dynamic_table_row_access_policy.*with row access policy always_true on \(id\)"
)

_MODELS_DYNAMIC_TABLE_TAG = (
    """
{{ config(
    target_lag = '7 days',
    snowflake_warehouse = '"""
    + os.getenv("SNOWFLAKE_TEST_WAREHOUSE", "")
    + """',
    materialized = 'dynamic_table',
    table_tag = "tag_name = 'tag_value'",
) }}

select * from {{ ref("my_seed") }}
"""
)
_DDL_MODELS_DYNAMIC_TABLE_TAG = r"\.dynamic_table_tag.*with tag \(tag_name = 'tag_value'\)"

_MODELS_DYNAMIC_TABLE_TAG_AND_ROW_ACCESS_POLICY = (
    """
{{ config(
    target_lag = '7 days',
    snowflake_warehouse = '"""
    + os.getenv("SNOWFLAKE_TEST_WAREHOUSE", "")
    + """',
    materialized = 'dynamic_table',
    table_tag = "tag_name = 'tag_value'",
    row_access_policy = 'always_true on (id)',
) }}

select * from {{ ref("my_seed") }}
"""
)
_DDL_MODELS_DYNAMIC_TABLE_TAG_AND_ROW_ACCESS_POLICY = r"\.dynamic_table_tag_and_row_access_policy.*with row access policy always_true on \(id\).*with tag \(tag_name = 'tag_value'\)"

_MODELS_ICEBERG_TABLE_TAG_AND_ROW_ACCESS_POLICY = """
{{
  config(
    transient = "true",
    materialized = "table",
    cluster_by=['id'],
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
    table_tag = "tag_name = 'tag_value'",
    row_access_policy = 'always_true on (id)',
  )
}}

select * from {{ ref('table_tag') }}
"""
_MODELS_ICEBERG_TABLE_TAG = """
{{
  config(
    transient = "true",
    materialized = "table",
    cluster_by=['id'],
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
    table_tag = "tag_name = 'tag_value'",
  )
}}

select * from {{ ref('table_tag') }}
"""

_MODELS_ICEBERG_TABLE_ROW_POLICY = """
{{
  config(
    transient = "true",
    materialized = "table",
    cluster_by=['id'],
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
    row_access_policy = 'always_true on (id)',
  )
}}

select * from {{ ref('table_tag') }}
"""
_DDL_MODELS_ICEBERG_TABLE_TAG = "with tag (tag_name = 'tag_value') "
_DDL_MODELS_ICEBERG_TABLE_POLICY = "with row access policy always_true on (id) "


class TestExtraConfig:
    @pytest.fixture(scope="class", autouse=True)
    def setup_policy(self, project):
        # Create the row access policy
        project.run_sql(_CREATE_TABLE_ROW_ACCESS_POLICY)
        project.run_sql(_CREATE_TABLE_TAG)

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "my_seed.csv": _MY_SEED,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            # Table models
            "table_row_access_policy.sql": _MODELS_TABLE_ROW_ACCESS_POLICY,
            "table_tag.sql": _MODELS_TABLE_TAG,
            "table_tag_and_row_access_policy.sql": _MODELS_TABLE_TAG_AND_ROW_ACCESS_POLICY,
            "table_iceberg_tag_and_row_access_policy.sql": _MODELS_ICEBERG_TABLE_TAG_AND_ROW_ACCESS_POLICY,
            "table_iceberg_tag.sql": _MODELS_ICEBERG_TABLE_TAG,
            "table_iceberg_row_policy.sql": _MODELS_ICEBERG_TABLE_ROW_POLICY,
            # View models
            "view_row_access_policy.sql": _MODELS_VIEW_ROW_ACCESS_POLICY,
            "view_tag.sql": _MODELS_VIEW_TAG,
            "view_tag_and_row_access_policy.sql": _MODELS_VIEW_TAG_AND_ROW_ACCESS_POLICY,
            # Incremental models
            "incremental_row_access_policy.sql": _MODELS_INCREMENTAL_ROW_ACCESS_POLICY,
            "incremental_tag.sql": _MODELS_INCREMENTAL_TAG,
            "incremental_tag_and_row_access_policy.sql": _MODELS_INCREMENTAL_TAG_AND_ROW_ACCESS_POLICY,
            # Dynamic table models
            "dynamic_table_row_access_policy.sql": _MODELS_DYNAMIC_TABLE_ROW_ACCESS_POLICY,
            "dynamic_table_tag.sql": _MODELS_DYNAMIC_TABLE_TAG,
            "dynamic_table_tag_and_row_access_policy.sql": _MODELS_DYNAMIC_TABLE_TAG_AND_ROW_ACCESS_POLICY,
        }

    def test_extra_config(self, project):
        run_dbt(["seed"])
        # depending on the Snowflake edition tags and row access policies are supported so we check the DDL sent
        results = run_dbt(["run"])
        assert len(results) == 15

        assert _DDL_MODELS_TABLE_ROW_ACCESS_POLICY in get_cleanded_model_ddl_from_file(
            "table_row_access_policy.sql"
        )
        iceberg_ddl_tag_and_policy = get_cleanded_model_ddl_from_file(
            "table_iceberg_tag_and_row_access_policy.sql"
        )
        iceberg_ddl_tag = get_cleanded_model_ddl_from_file("table_iceberg_tag.sql")
        iceberg_ddl_policy = get_cleanded_model_ddl_from_file("table_iceberg_row_policy.sql")
        assert all(
            _DDL_MODELS_ICEBERG_TABLE_POLICY in s
            for s in [iceberg_ddl_tag_and_policy, iceberg_ddl_policy]
        )
        assert all(
            _DDL_MODELS_ICEBERG_TABLE_TAG in s
            for s in [iceberg_ddl_tag_and_policy, iceberg_ddl_tag]
        )
        assert _DDL_MODELS_TABLE_TAG in get_cleanded_model_ddl_from_file("table_tag.sql")
        assert _DDL_MODELS_TABLE_TAG_AND_ROW_ACCESS_POLICY in get_cleanded_model_ddl_from_file(
            "table_tag_and_row_access_policy.sql"
        )
        assert _DDL_MODELS_VIEW_ROW_ACCESS_POLICY in get_cleanded_model_ddl_from_file(
            "view_row_access_policy.sql"
        )
        assert _DDL_MODELS_VIEW_TAG in get_cleanded_model_ddl_from_file("view_tag.sql")
        assert _DDL_MODELS_VIEW_TAG_AND_ROW_ACCESS_POLICY in get_cleanded_model_ddl_from_file(
            "view_tag_and_row_access_policy.sql"
        )

        assert _DDL_MODELS_INCREMENTAL_ROW_ACCESS_POLICY in get_cleanded_model_ddl_from_file(
            "incremental_row_access_policy.sql"
        )
        assert _DDL_MODELS_INCREMENTAL_TAG in get_cleanded_model_ddl_from_file(
            "incremental_tag.sql"
        )
        assert (
            _DDL_MODELS_INCREMENTAL_TAG_AND_ROW_ACCESS_POLICY
            in get_cleanded_model_ddl_from_file("incremental_tag_and_row_access_policy.sql")
        )

        assert re.search(
            _DDL_MODELS_DYNAMIC_TABLE_ROW_ACCESS_POLICY,
            get_cleanded_model_ddl_from_file("dynamic_table_row_access_policy.sql"),
        )
        assert re.search(
            _DDL_MODELS_DYNAMIC_TABLE_TAG,
            get_cleanded_model_ddl_from_file("dynamic_table_tag.sql"),
        )
        assert re.search(
            _DDL_MODELS_DYNAMIC_TABLE_TAG_AND_ROW_ACCESS_POLICY,
            get_cleanded_model_ddl_from_file("dynamic_table_tag_and_row_access_policy.sql"),
        )

    def test_extra_config_dynamic_table_full_refresh(self, project):
        # Test dynamic table models with row access policy and tags using full refresh
        run_dbt(["seed"])
        results = run_dbt(["run", "--select", "dynamic_table_*", "--full-refresh"])
        assert len(results) == 3

        assert re.search(
            _DDL_MODELS_DYNAMIC_TABLE_ROW_ACCESS_POLICY,
            get_cleanded_model_ddl_from_file("dynamic_table_row_access_policy.sql"),
        )
        assert re.search(
            _DDL_MODELS_DYNAMIC_TABLE_TAG,
            get_cleanded_model_ddl_from_file("dynamic_table_tag.sql"),
        )
        assert re.search(
            _DDL_MODELS_DYNAMIC_TABLE_TAG_AND_ROW_ACCESS_POLICY,
            get_cleanded_model_ddl_from_file("dynamic_table_tag_and_row_access_policy.sql"),
        )



from dbt.adapters.snowflake.relation import SnowflakeRelation
from dbt.adapters.snowflake.relation_configs import SnowflakeQuotePolicy


def test_relation_as_case_sensitive_quoting_true():
    relation = SnowflakeRelation.create(
        database="My_Db",
        schema="My_ScHeMa",
        identifier="My_TaBlE",
        quote_policy=SnowflakeQuotePolicy(database=False, schema=True, identifier=False),
    )

    case_sensitive_relation = relation.as_case_sensitive()
    case_sensitive_relation.render_limited()

    assert case_sensitive_relation.database == "MY_DB"
    assert case_sensitive_relation.schema == "My_ScHeMa"
    assert case_sensitive_relation.identifier == "MY_TABLE"
    assert case_sensitive_relation.render() == 'MY_DB."My_ScHeMa".MY_TABLE'



from types import SimpleNamespace

ADAPTER_TYPE = "snowflake"

DEFAULT_PYTHON_VERSION_FOR_PYTHON_MODELS = "3.10"

INFO_SCHEMA_TABLE_FORMAT = "DEFAULT"
ICEBERG_TABLE_FORMAT = "ICEBERG"

ICEBERG_REST_CATALOG_TYPE = "ICEBERG_REST"

DEFAULT_INFO_SCHEMA_CATALOG = SimpleNamespace(
    name="INFO_SCHEMA",  # these don't show up in Snowflake; this is a dbt convention
    catalog_type="INFO_SCHEMA",  # these don't show up in Snowflake; this is a dbt convention
    external_volume=None,
    file_format=None,
    adapter_properties={},
)
# catalog names must be uppercase since Snowflake will uppercase them in their metadata tables
DEFAULT_BUILT_IN_CATALOG = SimpleNamespace(
    name="SNOWFLAKE",
    catalog_type="BUILT_IN",
    # assume you are using the default volume or specify in the model
    external_volume=None,
    file_format=None,
    adapter_properties={},
)

DEFAULT_ICEBERG_REST_CATALOG = SimpleNamespace(
    name="POLARIS",
    catalog_type=ICEBERG_REST_CATALOG_TYPE,
    external_volume=None,
    table_format=ICEBERG_TABLE_FORMAT,
    file_format=None,
    adapter_properties={},
)

SnowflakeIcebergTableRelationParameters = SimpleNamespace(
    storage_serialization_policy="storage_serialization_policy",
    data_retention_time_in_days="data_retention_time_in_days",
    max_data_extension_time_in_days="max_data_extension_time_in_days",
    change_tracking="change_tracking",
    automatic_clustering="automatic_clustering",
)



import agate
import re
import unittest
from multiprocessing import get_context
from contextlib import contextmanager
from unittest import mock

from dbt.adapters.snowflake import SnowflakeAdapter
from dbt.adapters.snowflake import Plugin as SnowflakePlugin
from dbt.adapters.snowflake import SnowflakeAdapterResponse
from dbt.adapters.snowflake.column import SnowflakeColumn
from dbt.adapters.snowflake.connections import SnowflakeCredentials, SnowflakeConnectionManager
from dbt.contracts.files import FileHash
from dbt.context.query_header import generate_query_header_context
from dbt.context.providers import generate_runtime_macro_context
from dbt.contracts.graph.manifest import ManifestStateCheck
from dbt_common.clients import agate_helper
from snowflake import connector as snowflake_connector

from .utils import (
    config_from_parts_or_dicts,
    inject_adapter,
    mock_connection,
    TestAdapterConversions,
    load_internal_manifest_macros,
)


class TestSnowflakeAdapter(unittest.TestCase):
    def setUp(self):
        profile_cfg = {
            "outputs": {
                "test": {
                    "type": "snowflake",
                    "account": "test_account",
                    "user": "test_user",
                    "database": "test_database",
                    "warehouse": "test_warehouse",
                    "schema": "public",
                },
            },
            "target": "test",
        }

        project_cfg = {
            "name": "X",
            "version": "0.1",
            "profile": "test",
            "project-root": "/tmp/dbt/does-not-exist",
            "quoting": {
                "identifier": False,
                "schema": True,
            },
            "query-comment": "dbt",
            "config-version": 2,
        }
        self.config = config_from_parts_or_dicts(project_cfg, profile_cfg)
        self.assertEqual(self.config.query_comment.comment, "dbt")
        self.assertEqual(self.config.query_comment.append, None)

        self.handle = mock.MagicMock(spec=snowflake_connector.SnowflakeConnection)
        self.cursor = self.handle.cursor.return_value
        self.cursor.stats = None  # Explicitly set to None to avoid MagicMock auto-creation
        self.mock_execute = self.cursor.execute
        self.mock_execute.return_value = mock.MagicMock(sfqid="42")
        self.patcher = mock.patch("dbt.adapters.snowflake.connections.snowflake.connector.connect")
        self.snowflake = self.patcher.start()
        self.snowflake.connect.cursor.return_value = mock.MagicMock(sfqid="42")

        # Create the Manifest.state_check patcher
        @mock.patch("dbt.parser.manifest.ManifestLoader.build_manifest_state_check")
        def _mock_state_check(self):
            all_projects = self.all_projects
            return ManifestStateCheck(
                vars_hash=FileHash.from_contents("vars"),
                project_hashes={name: FileHash.from_contents(name) for name in all_projects},
                profile_hash=FileHash.from_contents("profile"),
            )

        self.load_state_check = mock.patch(
            "dbt.parser.manifest.ManifestLoader.build_manifest_state_check"
        )
        self.mock_state_check = self.load_state_check.start()
        self.mock_state_check.side_effect = _mock_state_check

        self.snowflake.return_value = self.handle
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        self.adapter.set_macro_resolver(load_internal_manifest_macros(self.config))
        self.adapter.set_macro_context_generator(generate_runtime_macro_context)
        self.adapter.connections.set_query_header(
            generate_query_header_context(self.config, self.adapter.get_macro_resolver())
        )

        self.qh_patch = mock.patch.object(self.adapter.connections.query_header, "add")
        self.mock_query_header_add = self.qh_patch.start()
        self.mock_query_header_add.side_effect = lambda q: "/* dbt */\n{}".format(q)
        self.adapter.acquire_connection()
        inject_adapter(self.adapter, SnowflakePlugin)

    def tearDown(self):
        # we want a unique self.handle every time.
        self.adapter.cleanup_connections()
        self.qh_patch.stop()
        self.patcher.stop()
        self.load_state_check.stop()

    def test_quoting_on_drop_schema(self):
        relation = SnowflakeAdapter.Relation.create(
            database="test_database",
            schema="test_schema",
            quote_policy=self.adapter.config.quoting,
        )
        self.adapter.drop_schema(relation)

        self.mock_execute.assert_has_calls(
            [
                mock.call(
                    '/* dbt */\ndrop schema if exists test_database."test_schema" cascade', None
                )
            ]
        )

    def test_quoting_on_drop(self):
        relation = self.adapter.Relation.create(
            database="test_database",
            schema="test_schema",
            identifier="test_table",
            type="table",
            quote_policy=self.adapter.config.quoting,
        )
        self.adapter.drop_relation(relation)

        self.mock_execute.assert_has_calls(
            [
                mock.call(
                    '/* dbt */\ndrop table if exists test_database."test_schema".test_table cascade',
                    None,
                )
            ]
        )

    def test_quoting_on_truncate(self):
        relation = self.adapter.Relation.create(
            database="test_database",
            schema="test_schema",
            identifier="test_table",
            type="table",
            quote_policy=self.adapter.config.quoting,
        )
        self.adapter.truncate_relation(relation)

        # no query comment because wrapped in begin; + commit; for explicit DML
        self.mock_execute.assert_has_calls(
            [
                mock.call("/* dbt */\nBEGIN", None),
                mock.call(
                    '/* dbt */\ntruncate table test_database."test_schema".test_table\n  ;', None
                ),
                mock.call("/* dbt */\nCOMMIT", None),
            ]
        )

    def test_quoting_on_rename(self):
        from_relation = self.adapter.Relation.create(
            database="test_database",
            schema="test_schema",
            identifier="table_a",
            type="table",
            quote_policy=self.adapter.config.quoting,
        )
        to_relation = self.adapter.Relation.create(
            database="test_database",
            schema="test_schema",
            identifier="table_b",
            type="table",
            quote_policy=self.adapter.config.quoting,
        )

        self.adapter.rename_relation(from_relation=from_relation, to_relation=to_relation)
        self.mock_execute.assert_has_calls(
            [
                mock.call(
                    '/* dbt */\nalter table test_database."test_schema".table_a rename to test_database."test_schema".table_b',
                    None,
                )
            ]
        )

    @contextmanager
    def current_warehouse(self, response):
        # there is probably some elegant way built into mock.patch to do this
        fetchall_return = self.cursor.fetchall.return_value
        execute_side_effect = self.mock_execute.side_effect

        def execute_effect(sql, *args, **kwargs):
            if sql == "/* dbt */\nselect current_warehouse() as warehouse":
                self.cursor.description = [["name"]]
                self.cursor.fetchall.return_value = [[response]]
            else:
                self.cursor.description = None
                self.cursor.fetchall.return_value = fetchall_return
            return self.mock_execute.return_value

        self.mock_execute.side_effect = execute_effect
        try:
            yield
        finally:
            self.cursor.fetchall.return_value = fetchall_return
            self.mock_execute.side_effect = execute_side_effect

    def _strip_transactions(self):
        result = []
        for call_args in self.mock_execute.call_args_list:
            args, kwargs = tuple(call_args)
            is_transactional = (
                len(kwargs) == 0
                and len(args) == 2
                and args[1] is None
                and args[0] in {"BEGIN", "COMMIT"}
            )
            if not is_transactional:
                result.append(call_args)
        return result

    def test_pre_post_hooks_warehouse(self):
        with self.current_warehouse("warehouse"):
            config = {"snowflake_warehouse": "other_warehouse"}
            result = self.adapter.pre_model_hook(config)
            self.assertIsNotNone(result)
            calls = [
                mock.call("/* dbt */\nselect current_warehouse() as warehouse", None),
                mock.call("/* dbt */\nuse warehouse other_warehouse", None),
            ]
            self.mock_execute.assert_has_calls(calls)
            self.adapter.post_model_hook(config, result)
            calls.append(mock.call("/* dbt */\nuse warehouse warehouse", None))
            self.mock_execute.assert_has_calls(calls)

    def test_pre_post_hooks_no_warehouse(self):
        with self.current_warehouse("warehouse"):
            config = {}
            result = self.adapter.pre_model_hook(config)
            self.assertIsNone(result)
            self.mock_execute.assert_not_called()
            self.adapter.post_model_hook(config, result)
            self.mock_execute.assert_not_called()

    def test_cancel_open_connections_empty(self):
        self.assertEqual(len(list(self.adapter.cancel_open_connections())), 0)

    def test_cancel_open_connections_master(self):
        key = self.adapter.connections.get_thread_identifier()
        self.adapter.connections.thread_connections[key] = mock_connection("master")
        self.assertEqual(len(list(self.adapter.cancel_open_connections())), 0)

    def test_cancel_open_connections_single(self):
        master = mock_connection("master")
        model = mock_connection("model")
        model.handle.session_id = 42

        key = self.adapter.connections.get_thread_identifier()
        self.adapter.connections.thread_connections.update(
            {
                key: master,
                1: model,
            }
        )
        with mock.patch.object(self.adapter.connections, "add_query") as add_query:
            query_result = mock.MagicMock()
            add_query.return_value = (None, query_result)

            self.assertEqual(len(list(self.adapter.cancel_open_connections())), 1)

            add_query.assert_called_once_with("select system$cancel_all_queries(42)")

    def test_client_session_keep_alive_false_by_default(self):
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    private_key=None,
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                ),
            ]
        )

    def test_client_session_keep_alive_true(self):
        self.config.credentials = self.config.credentials.replace(
            client_session_keep_alive=True,
            # this gets defaulted via `__post_init__` when `client_session_keep_alive` comes in as `False`
            # then when `replace` is called, `__post_init__` cannot set it back to `None` since it cannot
            # tell the difference between set by user and set by `__post_init__`
            reuse_connections=None,
        )
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=True,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    private_key=None,
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=None,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    def test_client_has_query_tag(self):
        self.config.credentials = self.config.credentials.replace(query_tag="test_query_tag")
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    reuse_connections=True,
                    warehouse="test_warehouse",
                    private_key=None,
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={"QUERY_TAG": "test_query_tag"},
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

        expected_connection_info = [
            (k, v) for (k, v) in self.config.credentials.connection_info() if k == "query_tag"
        ]
        self.assertEqual([("query_tag", "test_query_tag")], expected_connection_info)

    def test_user_pass_authentication(self):
        self.config.credentials = self.config.credentials.replace(
            password="test_password",
        )
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    password="test_password",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    private_key=None,
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    def test_authenticator_user_pass_authentication(self):
        self.config.credentials = self.config.credentials.replace(
            password="test_password",
            authenticator="test_sso_url",
        )
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    password="test_password",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    authenticator="test_sso_url",
                    private_key=None,
                    application="dbt",
                    client_request_mfa_token=True,
                    client_store_temporary_credential=True,
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    def test_authenticator_externalbrowser_authentication(self):
        self.config.credentials = self.config.credentials.replace(authenticator="externalbrowser")
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    authenticator="externalbrowser",
                    private_key=None,
                    application="dbt",
                    client_request_mfa_token=True,
                    client_store_temporary_credential=True,
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    def test_authenticator_oauth_authentication(self):
        self.config.credentials = self.config.credentials.replace(
            authenticator="oauth",
            token="my-oauth-token",
        )
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    authenticator="oauth",
                    token="my-oauth-token",
                    private_key=None,
                    application="dbt",
                    client_request_mfa_token=True,
                    client_store_temporary_credential=True,
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    @mock.patch(
        "dbt.adapters.snowflake.SnowflakeCredentials._get_private_key", return_value="test_key"
    )
    def test_authenticator_private_key_authentication(self, mock_get_private_key):
        self.config.credentials = self.config.credentials.replace(
            private_key_path="/tmp/test_key.p8",
            private_key_passphrase="p@ssphr@se",
        )

        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    private_key="test_key",
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    @mock.patch(
        "dbt.adapters.snowflake.SnowflakeCredentials._get_private_key", return_value="test_key"
    )
    def test_authenticator_private_key_authentication_no_passphrase(self, mock_get_private_key):
        self.config.credentials = self.config.credentials.replace(
            private_key_path="/tmp/test_key.p8",
            private_key_passphrase=None,
        )

        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    private_key="test_key",
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    def test_authenticator_jwt_authentication(self):
        self.config.credentials = self.config.credentials.replace(
            authenticator="jwt", token="my-jwt-token", user=None
        )
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    role=None,
                    schema="public",
                    warehouse="test_warehouse",
                    authenticator="oauth",
                    token="my-jwt-token",
                    private_key=None,
                    application="dbt",
                    client_request_mfa_token=True,
                    client_store_temporary_credential=True,
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    def test_query_tag(self):
        self.config.credentials = self.config.credentials.replace(
            password="test_password", query_tag="test_query_tag"
        )
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    password="test_password",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    private_key=None,
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={"QUERY_TAG": "test_query_tag"},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    def test_reuse_connections_with_keep_alive(self):
        self.config.credentials = self.config.credentials.replace(
            reuse_connections=True, client_session_keep_alive=True
        )
        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=True,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    private_key=None,
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    @mock.patch(
        "dbt.adapters.snowflake.SnowflakeCredentials._get_private_key", return_value="test_key"
    )
    def test_authenticator_private_key_string_authentication(self, mock_get_private_key):
        self.config.credentials = self.config.credentials.replace(
            private_key="dGVzdF9rZXk=",
            private_key_passphrase="p@ssphr@se",
        )

        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    private_key="test_key",
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    @mock.patch(
        "dbt.adapters.snowflake.SnowflakeCredentials._get_private_key", return_value="test_key"
    )
    def test_authenticator_private_key_string_authentication_no_passphrase(
        self, mock_get_private_key
    ):
        self.config.credentials = self.config.credentials.replace(
            private_key="dGVzdF9rZXk=",
            private_key_passphrase=None,
        )

        self.adapter = SnowflakeAdapter(self.config, get_context("spawn"))
        conn = self.adapter.connections.set_connection_name(name="new_connection_with_new_config")

        self.snowflake.assert_not_called()
        conn.handle
        self.snowflake.assert_has_calls(
            [
                mock.call(
                    account="test-account",
                    autocommit=True,
                    client_session_keep_alive=False,
                    database="test_database",
                    role=None,
                    schema="public",
                    user="test_user",
                    warehouse="test_warehouse",
                    private_key="test_key",
                    application="dbt",
                    insecure_mode=False,
                    platform_detection_timeout_seconds=0.0,
                    session_parameters={},
                    reuse_connections=True,
                    ocsp_root_certs_dict_lock_timeout=10,
                )
            ]
        )

    def test_get_relation_without_quotes(self):
        with mock.patch.object(self.adapter, "list_relations") as list_relations:
            list_relations.return_value = [
                SnowflakeAdapter.Relation.create(
                    database="TEST_DATABASE", schema="test_schema", identifier="TEST_TABLE"
                )
            ]
            relation = self.adapter.get_relation("test_database", "test_schema", "test_table")
            assert relation.render() == "TEST_DATABASE.test_schema.TEST_TABLE"

    def test_get_relation_with_quotes(self):
        with mock.patch.object(self.adapter, "list_relations") as list_relations:
            list_relations.return_value = [
                SnowflakeAdapter.Relation.create(
                    database="test_database", schema="test_schema", identifier="test_TABLE"
                )
            ]
            relation = self.adapter.get_relation('"test_database"', "test_schema", '"test_TABLE"')
            assert relation.render() == "test_database.test_schema.test_TABLE"


class TestSnowflakeAdapterConversions(TestAdapterConversions):
    def test_convert_text_type(self):
        rows = [
            ["", "a1", "stringval1"],
            ["", "a2", "stringvalasdfasdfasdfa"],
            ["", "a3", "stringval3"],
        ]
        agate_table = self._make_table_of(rows, agate.Text)
        expected = ["text", "text", "text"]
        for col_idx, expect in enumerate(expected):
            assert SnowflakeAdapter.convert_text_type(agate_table, col_idx) == expect

    def test_convert_number_type(self):
        rows = [
            ["", "23.98", "-1"],
            ["", "12.78", "-2"],
            ["", "79.41", "-3"],
        ]
        agate_table = self._make_table_of(rows, agate.Number)
        expected = ["integer", "float8", "integer"]
        for col_idx, expect in enumerate(expected):
            assert SnowflakeAdapter.convert_number_type(agate_table, col_idx) == expect

    def test_convert_boolean_type(self):
        rows = [
            ["", "false", "true"],
            ["", "false", "false"],
            ["", "false", "true"],
        ]
        agate_table = self._make_table_of(rows, agate.Boolean)
        expected = ["boolean", "boolean", "boolean"]
        for col_idx, expect in enumerate(expected):
            assert SnowflakeAdapter.convert_boolean_type(agate_table, col_idx) == expect

    def test_convert_datetime_type(self):
        rows = [
            ["", "20190101T01:01:01Z", "2019-01-01 01:01:01"],
            ["", "20190102T01:01:01Z", "2019-01-01 01:01:01"],
            ["", "20190103T01:01:01Z", "2019-01-01 01:01:01"],
        ]
        agate_table = self._make_table_of(
            rows, [agate.DateTime, agate_helper.ISODateTime, agate.DateTime]
        )
        expected = [
            "timestamp without time zone",
            "timestamp without time zone",
            "timestamp without time zone",
        ]
        for col_idx, expect in enumerate(expected):
            assert SnowflakeAdapter.convert_datetime_type(agate_table, col_idx) == expect

    def test_convert_date_type(self):
        rows = [
            ["", "2019-01-01", "2019-01-04"],
            ["", "2019-01-02", "2019-01-04"],
            ["", "2019-01-03", "2019-01-04"],
        ]
        agate_table = self._make_table_of(rows, agate.Date)
        expected = ["date", "date", "date"]
        for col_idx, expect in enumerate(expected):
            assert SnowflakeAdapter.convert_date_type(agate_table, col_idx) == expect

    def test_convert_time_type(self):
        # dbt's default type testers actually don't have a TimeDelta at all.
        agate.TimeDelta
        rows = [
            ["", "120s", "10s"],
            ["", "3m", "11s"],
            ["", "1h", "12s"],
        ]
        agate_table = self._make_table_of(rows, agate.TimeDelta)
        expected = ["time", "time", "time"]
        for col_idx, expect in enumerate(expected):
            assert SnowflakeAdapter.convert_time_type(agate_table, col_idx) == expect


class TestSnowflakeColumn(unittest.TestCase):
    def test_text_from_description(self):
        col = SnowflakeColumn.from_description("my_col", "TEXT")
        assert col.column == "my_col"
        assert col.dtype == "TEXT"
        assert col.char_size is None
        assert col.numeric_precision is None
        assert col.numeric_scale is None
        assert col.is_float() is False
        assert col.is_number() is False
        assert col.is_numeric() is False
        assert col.is_string() is True
        assert col.is_integer() is False
        assert col.string_size() == 16777216

        col = SnowflakeColumn.from_description("my_col", "VARCHAR")
        assert col.column == "my_col"
        assert col.dtype == "VARCHAR"
        assert col.char_size is None
        assert col.numeric_precision is None
        assert col.numeric_scale is None
        assert col.is_float() is False
        assert col.is_number() is False
        assert col.is_numeric() is False
        assert col.is_string() is True
        assert col.is_integer() is False
        assert col.string_size() == 16777216

    def test_sized_varchar_from_description(self):
        col = SnowflakeColumn.from_description("my_col", "VARCHAR(256)")
        assert col.column == "my_col"
        assert col.dtype == "VARCHAR"
        assert col.char_size == 256
        assert col.numeric_precision is None
        assert col.numeric_scale is None
        assert col.is_float() is False
        assert col.is_number() is False
        assert col.is_numeric() is False
        assert col.is_string() is True
        assert col.is_integer() is False
        assert col.string_size() == 256

    def test_sized_decimal_from_description(self):
        col = SnowflakeColumn.from_description("my_col", "DECIMAL(1, 0)")
        assert col.column == "my_col"
        assert col.dtype == "DECIMAL"
        assert col.char_size is None
        assert col.numeric_precision == 1
        assert col.numeric_scale == 0
        assert col.is_float() is False
        assert col.is_number() is True
        assert col.is_numeric() is True
        assert col.is_string() is False
        assert col.is_integer() is False

    def test_float_from_description(self):
        col = SnowflakeColumn.from_description("my_col", "FLOAT8")
        assert col.column == "my_col"
        assert col.dtype == "FLOAT8"
        assert col.char_size is None
        assert col.numeric_precision is None
        assert col.numeric_scale is None
        assert col.is_float() is True
        assert col.is_number() is True
        assert col.is_numeric() is False
        assert col.is_string() is False
        assert col.is_integer() is False

    def test_vector_from_description(self):
        col = SnowflakeColumn.from_description("my_col", "VECTOR(FLOAT, 768)")
        assert col.column == "my_col"
        assert col.dtype == "VECTOR(FLOAT, 768)"
        assert col.char_size is None
        assert col.numeric_precision is None
        assert col.numeric_scale is None
        assert col.is_float() is False
        assert col.is_number() is False
        assert col.is_numeric() is False
        assert col.is_string() is False
        assert col.is_integer() is False


class TestSnowflakeColumnStructuredTypes(unittest.TestCase):
    """
    https://docs.snowflake.com/en/sql-reference/data-types-structured

    Structured types would require a complete AST to fully parse through these types.
    For now, we opt for treating these as opaque strings of the type TYPE(INFO)
    where INFO could be types, sizes, or configuration params.
    """

    def test_array_from_description(self):
        """Documented example of optional NOT NULL param in the array type declaration"""
        col = SnowflakeColumn.from_description("my_col", "ARRAY(VARCHAR(16777216) NOT NULL)")
        assert col.column == "my_col"
        assert col.dtype == "ARRAY(VARCHAR(16777216) NOT NULL)"
        assert col.char_size is None
        assert col.numeric_precision is None
        assert col.numeric_scale is None
        assert col.is_float() is False
        assert col.is_number() is False
        assert col.is_numeric() is False
        assert col.is_string() is False
        assert col.is_integer() is False
        assert col.is_array() is True
        assert col.is_object() is False
        assert col.is_map() is False

    def test_object_from_description_iceberg(self):
        """Iceberg tables allow for extended type descriptions in OBJECT types"""
        col = SnowflakeColumn.from_description(
            "my_col",
            "OBJECT(languagePreference VARCHAR(16777216), localePreference VARCHAR(16777216))",
        )
        assert col.column == "my_col"
        assert (
            col.dtype
            == "OBJECT(languagePreference VARCHAR(16777216), localePreference VARCHAR(16777216))"
        )
        assert col.char_size is None
        assert col.numeric_precision is None
        assert col.numeric_scale is None
        assert col.is_float() is False
        assert col.is_number() is False
        assert col.is_numeric() is False
        assert col.is_string() is False
        assert col.is_integer() is False
        assert col.is_object() is True
        assert col.is_array() is False
        assert col.is_map() is False

    def test_object_from_description_simple(self):
        col = SnowflakeColumn.from_description("my_col", "OBJECT")
        assert col.column == "my_col"
        assert col.dtype == "OBJECT"
        assert col.char_size is None
        assert col.numeric_precision is None
        assert col.numeric_scale is None
        assert col.is_float() is False
        assert col.is_number() is False
        assert col.is_numeric() is False
        assert col.is_string() is False
        assert col.is_integer() is False
        assert col.is_object() is True
        assert col.is_array() is False
        assert col.is_map() is False

    def test_map_from_description(self):
        col = SnowflakeColumn.from_description("my_col", "MAP(VARCHAR(2), VARCHAR(2))")
        assert col.column == "my_col"
        assert col.dtype == "MAP(VARCHAR(2), VARCHAR(2))"
        assert col.char_size is None
        assert col.numeric_precision is None
        assert col.numeric_scale is None
        assert col.is_float() is False
        assert col.is_number() is False
        assert col.is_numeric() is False
        assert col.is_string() is False
        assert col.is_integer() is False
        assert col.is_map() is True
        assert col.is_object() is False
        assert col.is_array() is False


class SnowflakeConnectionsTest(unittest.TestCase):
    def test_comment_stripping_regex(self):
        pattern = r"(\".*?\"|\'.*?\')|(/\*.*?\*/|--[^\r\n]*$)"
        comment1 = "-- just comment"
        comment2 = "/* just comment */"
        query1 = "select 1; -- comment"
        query2 = "select 1; /* comment */"
        query3 = "select 1; -- comment\nselect 2; /* comment */ "
        query4 = "select \n1; -- comment\nselect \n2; /* comment */ "
        query5 = "select 1; -- comment \nselect 2; -- comment \nselect 3; -- comment"

        stripped_comment1 = re.sub(re.compile(pattern, re.MULTILINE), "", comment1).strip()

        stripped_comment2 = re.sub(re.compile(pattern, re.MULTILINE), "", comment2).strip()

        stripped_query1 = re.sub(re.compile(pattern, re.MULTILINE), "", query1).strip()

        stripped_query2 = re.sub(re.compile(pattern, re.MULTILINE), "", query2).strip()

        stripped_query3 = re.sub(re.compile(pattern, re.MULTILINE), "", query3).strip()

        stripped_query4 = re.sub(re.compile(pattern, re.MULTILINE), "", query4).strip()

        stripped_query5 = re.sub(re.compile(pattern, re.MULTILINE), "", query5).strip()

        expected_query_3 = "select 1; \nselect 2;"
        expected_query_4 = "select \n1; \nselect \n2;"
        expected_query_5 = "select 1; \nselect 2; \nselect 3;"

        self.assertEqual("", stripped_comment1)
        self.assertEqual("", stripped_comment2)
        self.assertEqual("select 1;", stripped_query1)
        self.assertEqual("select 1;", stripped_query2)
        self.assertEqual(expected_query_3, stripped_query3)
        self.assertEqual(expected_query_4, stripped_query4)
        self.assertEqual(expected_query_5, stripped_query5)


class TestSnowflakeAdapterCredentials(unittest.TestCase):
    unencrypted_private_key = (
        b"0\x82\x01T\x02\x01\x000\r\x06\t*\x86H\x86\xf7\r\x01\x01\x01"
        b"\x05\x00\x04\x82\x01>0\x82\x01:\x02\x01\x00\x02A\x00\xd9."
        b"\x15\xc8\xce\xfa\x1c\x9a\xe8/|5lf\xb8\xd9\x13\xc5I\x16i \x9f"
        b"'rO\xb1RkD(\n\xff\x84v\xbaS\x8d\xb46\xf8\x85w\x81\xe2\xc5cy"
        b"\xf1\xb6\xa9i]F\xfc\x04e`\xfbw;\x91\xf5\xcf\x02\x03\x01\x00"
        b"\x01\x02A\x00\x81\x84\xc6a\x17ny\x98\xb8WyO\xb2\xf2\x1f\xd2"
        b"\xf5\xc3v.\xf3K\r\x1fM@\xd1\x93A}H\x13\r\xa7\xd4\n,7L\x14?"
        b"\xff\xe2\xf3\xac\x93\xbb\xdf\xc3\xe5\xea\xf1AG\xc0~\xa2\x9a6"
        b"6\xeb\x11S\xe1\x02!\x00\xf3\x1d\xf1\xcc\xecj\xaf}\x01\xd4"
        b"\xee\x84\x03(Qx9\x9f\xedH\xf1\x016r\xbaE{Uk\x9d,\x13\x02!"
        b"\x00\xe4\xb0I\x8b\x8bU\xe5\x9a\x93V\x9f\xa8Ui\x9cQ\xd7\x12"
        b"\x866g<MK\xa4J5\xb0\xc6\xf7\xce\xd5\x02 wp=\xab\xe4v!R\xf3"
        b"\xc4m\x8d\x93\x93\x8a:\xdbl\x93\x81\xa3Mj7\x81\x05\xc3\xaa"
        b"\xda\x9c\xb3\xdb\x02 w\xafW^C\xd6\xf9\xaasp\x03p\xfa\xfa\xa1"
        b"\xc8'2W\xb1\x83H\t\x00\x0c\x84\x96\"\xe5\x8e\xed\xdd\x02 "
        b'\x1d\\\x02\xbc\xf3\xff\x17":>\xdf[{y\xe8\xc1j\xda\xc2\xa9'
        b"\x9d\x94)_\xe6\x9c\xf1FF\x03Y\x81"
    )
    unencrypted_private_key_encoded = (
        "MIIBVAIBADANBgkqhkiG9w0BAQEFAASCAT4wggE6AgEAAkEA2S4VyM76HJroL3w1bGa42"
        "RPFSRZpIJ8nck+xUmtEKAr/hHa6U420NviFd4HixWN58bapaV1G/ARlYPt3O5H1zwIDAQ"
        "ABAkEAgYTGYRdueZi4V3lPsvIf0vXDdi7zSw0fTUDRk0F9SBMNp9QKLDdMFD//4vOsk7v"
        "fw+Xq8UFHwH6imjY26xFT4QIhAPMd8czsaq99AdTuhAMoUXg5n+1I8QE2crpFe1VrnSwT"
        "AiEA5LBJi4tV5ZqTVp+oVWmcUdcShjZnPE1LpEo1sMb3ztUCIHdwPavkdiFS88RtjZOTi"
        "jrbbJOBo01qN4EFw6ranLPbAiB3r1deQ9b5qnNwA3D6+qHIJzJXsYNICQAMhJYi5Y7t3Q"
        "IgHVwCvPP/FyI6Pt9be3nowWrawqmdlClf5pzxRkYDWYE="
    )
    encrypted_private_key_encoded = (
        "MIIBvTBXBgkqhkiG9w0BBQ0wSjApBgkqhkiG9w0BBQwwHAQIIxdFbtlFbgkCAggAMAwGC"
        "CqGSIb3DQIJBQAwHQYJYIZIAWUDBAEqBBBswzZouyovzUADDQcLUEwgBIIBYAiB+rnGhm"
        "PwKZYOFdyEvkXFFu2aRfqotYHy/qlfVdU4BfNHwBlAlgUOPMN2HJ9KwyiNdBKoQ1Z4KXI"
        "G4AU74QZsVSL+miFf65qqWKLwckL45Y3WMUH1K0YpdO0W+aznjH9msWYuM/zCpQS2rvVX"
        "a5rXpA5praB5nn6kRlTwrQ8DN0ZKOKBX6ojhSE/6TmQtx3d+tmly8ZpTkG5HVTuBMCtDg"
        "Po6mAEvvb4T/dnx9MtUz0d5AgNuVOS5+OI32kX9XVupEGkvdY8iHbx7+skbKFVdHMayBL"
        "0dy5knySv+YGi/T6oM0uApPPk4aT493MNzT8544Wmi/NbNkDx+6XQiTuPZ4OsL0iF3KsX"
        "xfTC4tDGGYn4yW8bnSz3K+lkXA9vyie37nW6ncu+aizT9TgD1q6jFm9u/1G61/Z96oHgz"
        "pVghkP6s3l23U/7qM2PC8CEu18nUDYEhrv6lOwr8EABHV0s="
    )
    encrypted_private_key_passphrase = "insecure"

    def test_private_key_string(self):
        creds = SnowflakeCredentials(
            account="test-account",
            user="test_user",
            database="test_database",
            schema="public",
            private_key=self.unencrypted_private_key_encoded,
        )
        self.assertEqual(creds.auth_args()["private_key"], self.unencrypted_private_key)

    def test_private_key_string_encrypted(self):
        creds = SnowflakeCredentials(
            account="test-account",
            user="test_user",
            database="test_database",
            schema="public",
            private_key=self.encrypted_private_key_encoded,
            private_key_passphrase=self.encrypted_private_key_passphrase,
        )
        self.assertEqual(creds.auth_args()["private_key"], self.unencrypted_private_key)

    def test_malformed_private_key_string(self):
        creds = SnowflakeCredentials(
            account="test-account",
            user="test_user",
            database="test_database",
            schema="public",
            private_key="dGVzdF9rZXk=",
        )
        self.assertRaises(ValueError, creds.auth_args)

    def test_invalid_private_key_string(self):
        creds = SnowflakeCredentials(
            account="test-account",
            user="test_user",
            database="test_database",
            schema="public",
            private_key="invalid[base64]=",
        )
        self.assertRaises(ValueError, creds.auth_args)

    def test_invalid_private_key_path(self):
        creds = SnowflakeCredentials(
            account="test-account",
            user="test_user",
            database="test_database",
            schema="public",
            private_key_path="/tmp/does/not/exist.p8",
        )
        self.assertRaises(FileNotFoundError, creds.auth_args)


class TestSnowflakeGetResponse(unittest.TestCase):
    """Tests for SnowflakeConnectionManager.get_response with cursor.stats support."""

    def test_get_response_with_stats(self):
        """Test get_response populates stats from cursor.stats for DML operations."""
        # Create a mock cursor with stats (simulating snowflake-connector-python >= 4.2.0)
        cursor = mock.MagicMock()
        cursor.sqlstate = None
        cursor.rowcount = 1
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"

        # Mock stats object (like QueryResultStats from snowflake connector)
        stats = mock.MagicMock()
        stats.num_rows_inserted = 100
        stats.num_rows_deleted = 0
        stats.num_rows_updated = 0
        stats.num_dml_duplicates = 0
        cursor.stats = stats

        response = SnowflakeConnectionManager.get_response(cursor)

        assert isinstance(response, SnowflakeAdapterResponse)
        assert response.code == "SUCCESS"
        # rows_affected uses rows_inserted when available and > 0
        assert response.rows_affected == 100
        assert response.query_id == "01abc123-0000-0000-0000-000000000000"
        assert response.rows_inserted == 100
        assert response.rows_deleted == 0
        assert response.rows_updated == 0
        assert response.rows_duplicates == 0

    def test_get_response_with_stats_none_values(self):
        """Test get_response when cursor.stats has None values (non-DML operations)."""
        cursor = mock.MagicMock()
        cursor.sqlstate = None
        cursor.rowcount = 1
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"

        # Mock stats object with None values (like for CREATE VIEW)
        stats = mock.MagicMock()
        stats.num_rows_inserted = None
        stats.num_rows_deleted = None
        stats.num_rows_updated = None
        stats.num_dml_duplicates = None
        cursor.stats = stats

        response = SnowflakeConnectionManager.get_response(cursor)

        assert isinstance(response, SnowflakeAdapterResponse)
        assert response.code == "SUCCESS"
        assert response.rows_affected == 1
        assert response.rows_inserted is None
        assert response.rows_deleted is None
        assert response.rows_updated is None
        assert response.rows_duplicates is None

    def test_get_response_without_stats_attribute(self):
        """Test get_response when cursor doesn't have stats attribute (older connector)."""
        cursor = mock.MagicMock(spec=["sqlstate", "rowcount", "sfqid"])
        cursor.sqlstate = None
        cursor.rowcount = 5
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"

        response = SnowflakeConnectionManager.get_response(cursor)

        assert isinstance(response, SnowflakeAdapterResponse)
        assert response.code == "SUCCESS"
        assert response.rows_affected == 5
        assert response.rows_inserted is None
        assert response.rows_deleted is None
        assert response.rows_updated is None
        assert response.rows_duplicates is None

    def test_get_response_with_stats_none(self):
        """Test get_response when cursor.stats is None."""
        cursor = mock.MagicMock()
        cursor.sqlstate = None
        cursor.rowcount = 3
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"
        cursor.stats = None

        response = SnowflakeConnectionManager.get_response(cursor)

        assert isinstance(response, SnowflakeAdapterResponse)
        assert response.code == "SUCCESS"
        assert response.rows_affected == 3
        assert response.rows_inserted is None
        assert response.rows_deleted is None
        assert response.rows_updated is None
        assert response.rows_duplicates is None

    def test_get_response_with_sqlstate(self):
        """Test get_response preserves sqlstate code when present."""
        cursor = mock.MagicMock()
        cursor.sqlstate = "02000"
        cursor.rowcount = 0
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"
        cursor.stats = None

        response = SnowflakeConnectionManager.get_response(cursor)

        assert isinstance(response, SnowflakeAdapterResponse)
        assert response.code == "02000"
        assert str(response) == "02000 0"

    def test_get_response_with_update_stats(self):
        """Test get_response with UPDATE operation stats."""
        cursor = mock.MagicMock()
        cursor.sqlstate = None
        cursor.rowcount = 50
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"

        stats = mock.MagicMock()
        stats.num_rows_inserted = 0
        stats.num_rows_deleted = 0
        stats.num_rows_updated = 50
        stats.num_dml_duplicates = 0
        cursor.stats = stats

        response = SnowflakeConnectionManager.get_response(cursor)

        assert response.rows_affected == 50
        assert response.rows_inserted == 0
        assert response.rows_updated == 50
        assert response.rows_deleted == 0

    def test_get_response_with_delete_stats(self):
        """Test get_response with DELETE operation stats."""
        cursor = mock.MagicMock()
        cursor.sqlstate = None
        cursor.rowcount = 25
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"

        stats = mock.MagicMock()
        stats.num_rows_inserted = 0
        stats.num_rows_deleted = 25
        stats.num_rows_updated = 0
        stats.num_dml_duplicates = 0
        cursor.stats = stats

        response = SnowflakeConnectionManager.get_response(cursor)

        assert response.rows_affected == 25
        assert response.rows_inserted == 0
        assert response.rows_deleted == 25
        assert response.rows_updated == 0

    def test_get_response_ctas_uses_rows_inserted(self):
        """Test get_response uses rows_inserted for rows_affected for CTAS operations."""
        cursor = mock.MagicMock()
        cursor.sqlstate = None
        cursor.rowcount = 1  # CTAS returns 1 for rowcount (the success message row)
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"

        stats = mock.MagicMock()
        stats.num_rows_inserted = 1000
        stats.num_rows_deleted = 0
        stats.num_rows_updated = 0
        stats.num_dml_duplicates = 0
        cursor.stats = stats

        response = SnowflakeConnectionManager.get_response(cursor)

        # rows_affected should use rows_inserted for accurate reporting
        assert response.rows_affected == 1000
        assert response.rows_inserted == 1000
        assert str(response) == "SUCCESS 1000"

    def test_get_response_uses_rowcount_when_no_rows_inserted(self):
        """Test get_response uses rowcount when rows_inserted is 0 or None."""
        cursor = mock.MagicMock()
        cursor.sqlstate = None
        cursor.rowcount = 1
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"

        stats = mock.MagicMock()
        stats.num_rows_inserted = 0  # No rows inserted
        stats.num_rows_deleted = 0
        stats.num_rows_updated = 0
        stats.num_dml_duplicates = 0
        cursor.stats = stats

        response = SnowflakeConnectionManager.get_response(cursor)

        # rows_affected should use rowcount when rows_inserted is 0
        assert response.rows_affected == 1
        assert response.rows_inserted == 0
        assert str(response) == "SUCCESS 1"

    def test_get_response_rowcount_when_no_stats(self):
        """Test get_response uses rowcount when no stats available."""
        cursor = mock.MagicMock()
        cursor.sqlstate = None
        cursor.rowcount = 5
        cursor.sfqid = "01abc123-0000-0000-0000-000000000000"
        cursor.stats = None

        response = SnowflakeConnectionManager.get_response(cursor)

        # rows_affected should use rowcount when no stats
        assert response.rows_affected == 5
        assert response.rows_inserted is None
        assert str(response) == "SUCCESS 5"



import pytest
from dbt.tests.adapter.incremental.test_incremental_unique_id import BaseIncrementalUniqueKey


class TestUniqueKeySnowflake(BaseIncrementalUniqueKey):
    pass


class TestUniqueKeyDeleteInsertSnowflake(BaseIncrementalUniqueKey):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"+incremental_strategy": "delete+insert"}}



import dataclasses
from typing import Optional

from dbt_common.record import Record, Recorder, record_function

from dbt.adapters.record import RecordReplayHandle
from dbt.adapters.snowflake.record.cursor.cursor import SnowflakeRecordReplayCursor


@dataclasses.dataclass
class HandleGetSessionIdParams:
    connection_name: str


@dataclasses.dataclass
class HandleGetSessionIdResult:
    session_id: Optional[int]


@Recorder.register_record_type
class HandleGetSessionIdRecord(Record):
    params_cls = HandleGetSessionIdParams
    result_cls = HandleGetSessionIdResult
    group = "Database"


class SnowflakeRecordReplayHandle(RecordReplayHandle):
    """A custom extension of RecordReplayHandle that returns a
    snowflake-connector-specific SnowflakeRecordReplayCursor object and adds
    the session_id property which is specific to snowflake-connector."""

    def cursor(self):
        cursor = None if self.native_handle is None else self.native_handle.cursor()
        return SnowflakeRecordReplayCursor(cursor, self.connection)

    @property
    def connection_name(self) -> Optional[str]:
        return self.connection.name

    @property
    @record_function(HandleGetSessionIdRecord, method=True, id_field_name="connection_name")
    def session_id(self):
        return self.native_handle.session_id



from dbt.adapters.snowflake.catalogs._built_in import (
    BuiltInCatalogIntegration,
    BuiltInCatalogRelation,
)
from dbt.adapters.snowflake.catalogs._info_schema import (
    InfoSchemaCatalogIntegration,
    InfoSchemaCatalogRelation,
)
from dbt.adapters.snowflake.catalogs._iceberg_rest import (
    IcebergRestCatalogIntegration,
    IcebergRestCatalogRelation,
)



from dbt.tests.adapter.utils.data_types.test_type_bigint import BaseTypeBigInt
from dbt.tests.adapter.utils.data_types.test_type_float import BaseTypeFloat
from dbt.tests.adapter.utils.data_types.test_type_int import BaseTypeInt
from dbt.tests.adapter.utils.data_types.test_type_numeric import BaseTypeNumeric
from dbt.tests.adapter.utils.data_types.test_type_string import BaseTypeString
from dbt.tests.adapter.utils.data_types.test_type_timestamp import BaseTypeTimestamp
from dbt.tests.adapter.utils.data_types.test_type_boolean import BaseTypeBoolean


class TestTypeBigInt(BaseTypeBigInt):
    pass


class TestTypeFloat(BaseTypeFloat):
    pass


class TestTypeInt(BaseTypeInt):
    pass


class TestTypeNumeric(BaseTypeNumeric):
    pass


class TestTypeString(BaseTypeString):
    pass


class TestTypeTimestamp(BaseTypeTimestamp):
    pass


class TestTypeBoolean(BaseTypeBoolean):
    pass



import os

import pytest

from dbt.tests.util import run_dbt


SEED = """
id
1
""".strip()


MODEL = """
{{ config(
    materialized='incremental',
) }}
select * from {{ ref('my_seed') }}
"""


class TestDatabaseRole:
    """
    This test addresses https://github.com/dbt-labs/dbt-snowflake/issues/1151

    While dbt-snowflake does not manage database roles (it only manages account roles,
    it still needs to account for them so that it doesn't try to revoke them.
    """

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"my_seed.csv": SEED}

    @pytest.fixture(scope="class")
    def models(self):
        return {"my_model.sql": MODEL}

    @pytest.fixture(scope="class")
    def project_config_update(self):
        # grant to the test role even though this role already has these permissions
        # this triggers syncing grants since `apply_grants` first looks for a grants config
        return {"models": {"+grants": {"select": [os.getenv("SNOWFLAKE_TEST_ROLE")]}}}

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project, prefix):
        """
        Create a database role with access to the model we're about to create.
        The existence of this database role triggered the bug as dbt-snowflake attempts
        to revoke it if the user also provides a grants config.
        """
        role = f"BLOCKING_DB_ROLE_{prefix}"
        project.run_sql(f"CREATE DATABASE ROLE {role}")
        sql = f"""
        GRANT
            ALL PRIVILEGES ON FUTURE TABLES
            IN SCHEMA {project.test_schema}
            TO DATABASE ROLE {role}
        """
        project.run_sql(sql)
        yield
        project.run_sql(f"DROP DATABASE ROLE {role}")

    def test_database_role(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])
        # run a second time to trigger revoke on an incremental update
        # this originally failed, demonstrating the bug
        run_dbt(["run"])



"""
The first time using an account for testing, you should run this:

```
CREATE OR REPLACE SECURITY INTEGRATION DBT_INTEGRATION_TEST_OAUTH
  TYPE = OAUTH
  ENABLED = TRUE
  OAUTH_CLIENT = CUSTOM
  OAUTH_CLIENT_TYPE = 'CONFIDENTIAL'
  OAUTH_REDIRECT_URI = 'http://localhost:8080'
  oauth_issue_refresh_tokens = true
  OAUTH_ALLOW_NON_TLS_REDIRECT_URI = true
  BLOCKED_ROLES_LIST = <everything but your integration test role goes here: ('ACCOUNTADMIN', 'SECURITYADMIN'), or ignore it if you don't care>
  oauth_refresh_token_validity = 7776000;
```


Every month (or any amount <90 days):

Run `select SYSTEM$SHOW_OAUTH_CLIENT_SECRETS('DBT_INTEGRATION_TEST_OAUTH');`

The only row/column of output should be a json blob, it goes (within single
quotes!) as the second argument to the server script:

python scripts/werkzeug-refresh-token.py ${acount_name} '${json_blob}'

Open http://localhost:8080

Log in as the test user, get a response page with some environment variables.
Update CI providers and test.env with the new values (If you kept the security
integration the same, just the refresh token changed)
"""

import os
from dbt.tests.util import check_relations_equal, run_dbt
import pytest


_MODELS__MODEL_1_SQL = """
select 1 as id
"""


_MODELS__MODEL_2_SQL = """
select 2 as id
"""


_MODELS__MODEL_3_SQL = """
select * from {{ ref('model_1') }}
union all
select * from {{ ref('model_2') }}
"""


_MODELS__MODEL_4_SQL = """
select 1 as id
union all
select 2 as id
"""


class TestSnowflakeOauth:
    @pytest.fixture(scope="class", autouse=True)
    def dbt_profile_target(self):
        return {
            "type": "snowflake",
            "threads": 4,
            "account": os.getenv("SNOWFLAKE_TEST_ACCOUNT"),
            "user": os.getenv("SNOWFLAKE_TEST_USER"),
            "oauth_client_id": os.getenv("SNOWFLAKE_TEST_OAUTH_CLIENT_ID"),
            "oauth_client_secret": os.getenv("SNOWFLAKE_TEST_OAUTH_CLIENT_SECRET"),
            "token": os.getenv("SNOWFLAKE_TEST_OAUTH_REFRESH_TOKEN"),
            "database": os.getenv("SNOWFLAKE_TEST_DATABASE"),
            "warehouse": os.getenv("SNOWFLAKE_TEST_WAREHOUSE"),
            "authenticator": "oauth",
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model_1.sql": _MODELS__MODEL_1_SQL,
            "model_2.sql": _MODELS__MODEL_2_SQL,
            "model_3.sql": _MODELS__MODEL_3_SQL,
            "model_4.sql": _MODELS__MODEL_4_SQL,
        }

    def test_snowflake_basic(self, project):
        run_dbt()
        check_relations_equal(project.adapter, ["MODEL_3", "MODEL_4"])



from dbt.tests.adapter.incremental.test_incremental_on_schema_change import (
    BaseIncrementalOnSchemaChange,
    BaseIncrementalOnSchemaChangeSetup,
)
from dbt.tests.util import run_dbt
import pytest


class TestIncrementalOnSchemaChange(BaseIncrementalOnSchemaChange):
    pass


# New fixtures for testing column names with spaces and special characters
_MODEL_A_SPECIAL_COLUMNS = """
{{
    config(materialized='table')
}}

with source_data as (

    select 1 as id, 'aaa' as "field with space", 'bbb' as "field-with-dash", 111 as "field@special", 'TTT' as "field.with.dots"
    union all select 2 as id, 'ccc' as "field with space", 'ddd' as "field-with-dash", 222 as "field@special", 'UUU' as "field.with.dots"
    union all select 3 as id, 'eee' as "field with space", 'fff' as "field-with-dash", 333 as "field@special", 'VVV' as "field.with.dots"
    union all select 4 as id, 'ggg' as "field with space", 'hhh' as "field-with-dash", 444 as "field@special", 'WWW' as "field.with.dots"
    union all select 5 as id, 'iii' as "field with space", 'jjj' as "field-with-dash", 555 as "field@special", 'XXX' as "field.with.dots"
    union all select 6 as id, 'kkk' as "field with space", 'lll' as "field-with-dash", 666 as "field@special", 'YYY' as "field.with.dots"

)

select id
       ,"field with space"
       ,"field-with-dash"
       ,"field@special"
       ,"field.with.dots"

from source_data
"""

_MODEL_INCREMENTAL_APPEND_NEW_COLUMNS_SPECIAL = """
{{
    config(
        materialized='incremental',
        unique_key='id',
        on_schema_change='append_new_columns'
    )
}}

WITH source_data AS (SELECT * FROM {{ ref('model_a_special_columns') }} )

{% if is_incremental() %}

SELECT id,
       "field with space",
       "field-with-dash",
       "field@special",
       "field.with.dots"
FROM source_data WHERE id NOT IN (SELECT id from {{ this }} )

{% else %}

SELECT id,
       "field with space",
       "field-with-dash"
FROM source_data where id <= 3

{% endif %}
"""

_MODEL_INCREMENTAL_APPEND_NEW_COLUMNS_SPECIAL_TARGET = """
{{
    config(materialized='table')
}}

with source_data as (

    select * from {{ ref('model_a_special_columns') }}

)

select id
       ,"field with space"
       ,"field-with-dash"
       ,CASE WHEN id <= 3 THEN NULL ELSE "field@special" END AS "field@special"
       ,CASE WHEN id <= 3 THEN NULL ELSE "field.with.dots" END AS "field.with.dots"

from source_data
"""


class TestIncrementalOnSchemaChangeSpecialColumns(BaseIncrementalOnSchemaChangeSetup):
    """Test incremental models with on_schema_change when column names contain spaces and special characters"""

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model_a_special_columns.sql": _MODEL_A_SPECIAL_COLUMNS,
            "incremental_append_new_columns_special.sql": _MODEL_INCREMENTAL_APPEND_NEW_COLUMNS_SPECIAL,
            "incremental_append_new_columns_special_target.sql": _MODEL_INCREMENTAL_APPEND_NEW_COLUMNS_SPECIAL_TARGET,
        }

    def test_incremental_append_new_columns_with_special_characters(self, project):
        """Test that incremental models with on_schema_change='append_new_columns' work correctly with column names containing spaces and special characters"""

        # First run - creates initial table with id, "field with space", "field-with-dash"
        run_dbt(
            [
                "run",
                "--models",
                "model_a_special_columns incremental_append_new_columns_special incremental_append_new_columns_special_target",
            ]
        )

        # Second run - should append new columns "field@special" and "field.with.dots"
        run_dbt(
            [
                "run",
                "--models",
                "model_a_special_columns incremental_append_new_columns_special incremental_append_new_columns_special_target",
            ]
        )

        # Verify that the incremental model matches the target
        from dbt.tests.util import check_relations_equal

        check_relations_equal(
            project.adapter,
            [
                "incremental_append_new_columns_special",
                "incremental_append_new_columns_special_target",
            ],
        )



import pytest
from dbt.tests.util import run_dbt_and_capture

my_model_sql = """
{{
    config(
        materialized = 'table',
        post_hook = '{{ my_silly_insert_macro() }}'
    )
}}
select 1 as id, 'blue' as color, current_timestamp as updated_at
"""

my_macro_sql = """
{% macro my_silly_insert_macro() %}
    {#-- This is a bad pattern! Made obsolete by changes in v0.21 + v1.2 --#}
    {% do run_query('begin;') %}
    {% set query %}
       insert into {{ this }} values (2, 'red', current_timestamp);
    {% endset %}
    {% do run_query(query) %}
    {% do run_query('commit;') %}
{% endmacro %}
"""


class TestModelWarehouse:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_sql,
        }

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "my_macro.sql": my_macro_sql,
        }

    def test_isolated_begin_commit(self, project):
        # this should succeed / avoid raising an error
        results, log_output = run_dbt_and_capture(["run"])
        # but we should see a warning in the logs
        assert "WARNING" in log_output and "Explicit transactional logic" in log_output



import pytest

from dbt.tests.adapter.unit_testing.test_types import BaseUnitTestingTypes
from dbt.tests.adapter.unit_testing.test_case_insensitivity import BaseUnitTestCaseInsensivity
from dbt.tests.adapter.unit_testing.test_invalid_input import BaseUnitTestInvalidInput


class TestSnowflakeUnitTestingTypes(BaseUnitTestingTypes):
    @pytest.fixture
    def data_types(self):
        # sql_value, yaml_value
        return [
            ["1", "1"],
            ["2.0", "2.0"],
            ["'12345'", "12345"],
            ["'string'", "string"],
            ["true", "true"],
            ["DATE '2020-01-02'", "2020-01-02"],
            ["TIMESTAMP '2013-11-03 00:00:00-0'", "2013-11-03 00:00:00-0"],
            ["'2013-11-03 00:00:00-0'::TIMESTAMPTZ", "2013-11-03 00:00:00-0"],
            ["TO_NUMBER('3', 10, 9)", "3"],
            ["3::VARIANT", "3"],
            ["TO_GEOMETRY('POINT(1820.12 890.56)')", "POINT(1820.12 890.56)"],
            ["TO_GEOGRAPHY('POINT(-122.35 37.55)')", "POINT(-122.35 37.55)"],
            [
                "{'Alberta':'Edmonton','Manitoba':'Winnipeg'}",
                "{'Alberta':'Edmonton','Manitoba':'Winnipeg'}",
            ],
            ["['a','b','c']", "['a','b','c']"],
            ["[1,2,3]", "[1, 2, 3]"],
        ]


class TestSnowflakeUnitTestCaseInsensitivity(BaseUnitTestCaseInsensivity):
    pass


class TestSnowflakeUnitTestInvalidInput(BaseUnitTestInvalidInput):
    pass



import pytest
from dbt.tests.adapter.column_types.test_column_types import BaseColumnTypes
from tests.functional.adapter.column_types.fixtures import _MODEL_SQL, _SCHEMA_YML


class TestSnowflakeColumnTypes(BaseColumnTypes):
    @pytest.fixture(scope="class")
    def models(self):
        return {"model.sql": _MODEL_SQL, "schema.yml": _SCHEMA_YML}

    def test_run_and_test(self, project):
        self.run_and_test()



from dbt.adapters.base.query_headers import MacroQueryStringSetter


class SnowflakeMacroQueryStringSetter(MacroQueryStringSetter):
    # Snowflake removes query headers that are prepended by default.
    # In order to persist them by default in dbt, they must be appended.
    DEFAULT_QUERY_COMMENT_APPEND = True



import pytest
import time

from pathlib import Path

from dbt.tests.util import run_dbt, run_dbt_and_capture, write_file


_SEED_INCREMENTAL_STRATEGIES = """
world_id,world_name,boss
1,Yoshi's Island,Iggy
2,Donut Plains,Morton
3,Vanilla Dome,Lemmy
4,Cookie Mountain,Temmy
5,Forest of Illusion,Roy
""".strip()

_MODEL_BASIC_TABLE_MODEL = """
{{
  config(
    materialized = "table",
  )
}}
select * from {{ ref('seed') }}
"""

_MODEL_INCREMENTAL_ICEBERG_BASE = """
{{{{
  config(
    materialized='incremental',
    table_format='iceberg',
    incremental_strategy='{strategy}',
    unique_key="world_id",
    external_volume = "s3_iceberg_snow",
    on_schema_change = "sync_all_columns"
  )
}}}}
select * from {{{{ ref('upstream_table') }}}}

{{% if is_incremental() %}}
where world_id > 2
{{% endif %}}
"""

_MODEL_INCREMENTAL_ICEBERG_BUILTIN_BASE = """
{{{{
  config(
    materialized='incremental',
    catalog='snowflake',
    incremental_strategy='{strategy}',
    unique_key="world_id",
    external_volume = "s3_iceberg_snow",
    on_schema_change = "sync_all_columns"
  )
}}}}
select * from {{{{ ref('upstream_table') }}}}

{{% if is_incremental() %}}
where world_id > 2
{{% endif %}}
"""


_MODEL_INCREMENTAL_ICEBERG_APPEND = _MODEL_INCREMENTAL_ICEBERG_BASE.format(strategy="append")
_MODEL_INCREMENTAL_ICEBERG_MERGE = _MODEL_INCREMENTAL_ICEBERG_BASE.format(strategy="merge")
_MODEL_INCREMENTAL_ICEBERG_DELETE_INSERT = _MODEL_INCREMENTAL_ICEBERG_BASE.format(
    strategy="delete+insert"
)


_MODEL_INCREMENTAL_ICEBERG_BUILTIN_APPEND = _MODEL_INCREMENTAL_ICEBERG_BUILTIN_BASE.format(
    strategy="append"
)
_MODEL_INCREMENTAL_ICEBERG_BUILTIN_MERGE = _MODEL_INCREMENTAL_ICEBERG_BUILTIN_BASE.format(
    strategy="merge"
)
_MODEL_INCREMENTAL_ICEBERG_BUILTIN_DELETE_INSERT = _MODEL_INCREMENTAL_ICEBERG_BUILTIN_BASE.format(
    strategy="delete+insert"
)


_QUERY_UPDATE_UPSTREAM_TABLE = """
UPDATE {database}.{schema}.upstream_table set world_name = 'Twin Bridges', boss = 'Ludwig' where world_id = 4;
"""

_QUERY_UPDATE_UPSTREAM_TABLE_NO_EFFECT = """
UPDATE {database}.{schema}.upstream_table set world_name = 'Doughnut Plains' where world_id = 2;
"""


class TestIcebergIncrementalStrategies:
    append: str = f"append_{hash(time.time())}"

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": _SEED_INCREMENTAL_STRATEGIES,
        }

    @pytest.fixture(scope="function", autouse=True)
    def setup_class(self, project):
        run_dbt(["seed"])
        yield

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "upstream_table.sql": _MODEL_BASIC_TABLE_MODEL,
            f"{self.append}.sql": _MODEL_INCREMENTAL_ICEBERG_APPEND,
            "merge.sql": _MODEL_INCREMENTAL_ICEBERG_MERGE,
            "delete_insert.sql": _MODEL_INCREMENTAL_ICEBERG_DELETE_INSERT,
        }

    def __check_correct_operations(self, model_name, /, rows_affected, status="SUCCESS"):
        run_results = run_dbt(
            ["show", "--inline", f"select * from {{{{ ref('{model_name}') }}}} where world_id = 4"]
        )
        assert run_results[0].adapter_response["rows_affected"] == rows_affected
        assert run_results[0].adapter_response["code"] == status

        if "append" not in model_name:
            run_results, stdout = run_dbt_and_capture(
                [
                    "show",
                    "--inline",
                    f"select * from {{{{ ref('{model_name}') }}}} where world_id = 2",
                ]
            )
            run_results[0].adapter_response["rows_affected"] == 1
            assert "Doughnut" not in stdout

    def test_incremental_strategies_with_update(self, project, setup_class):
        run_results = run_dbt()
        assert len(run_results) == 4

        project.run_sql(
            _QUERY_UPDATE_UPSTREAM_TABLE.format(
                database=project.database, schema=project.test_schema
            )
        )
        project.run_sql(
            _QUERY_UPDATE_UPSTREAM_TABLE_NO_EFFECT.format(
                database=project.database, schema=project.test_schema
            )
        )

        run_results = run_dbt(["run", "-s", self.append, "merge", "delete_insert"])
        assert len(run_results) == 3

        self.__check_correct_operations(self.append, rows_affected=2)
        self.__check_correct_operations("merge", rows_affected=1)
        self.__check_correct_operations("delete_insert", rows_affected=1)


class TestIcebergBuiltInIncrementalStrategies(TestIcebergIncrementalStrategies):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "upstream_table.sql": _MODEL_BASIC_TABLE_MODEL,
            f"{self.append}.sql": _MODEL_INCREMENTAL_ICEBERG_BUILTIN_APPEND,
            "merge.sql": _MODEL_INCREMENTAL_ICEBERG_BUILTIN_MERGE,
            "delete_insert.sql": _MODEL_INCREMENTAL_ICEBERG_BUILTIN_DELETE_INSERT,
        }


class TestIcebergIncrementalOnSchemaChangeMutatesRelations:

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": _SEED_INCREMENTAL_STRATEGIES,
        }

    @pytest.fixture(scope="function", autouse=True)
    def setup_class(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])
        yield

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "upstream_table.sql": _MODEL_BASIC_TABLE_MODEL,
            "merge.sql": _MODEL_INCREMENTAL_ICEBERG_MERGE,
        }

    def test_sync_and_append_semantics(self, project, setup_class):
        model_file = project.project_root / Path("models") / Path("merge.sql")
        sql = f"show columns in {project.database}.{project.test_schema}.merge;"
        column_names = [column[2] for column in project.run_sql(sql, fetch="all")]
        assert len(column_names) == 3

        write_file(_MODEL_INCREMENTAL_ICEBERG_MERGE.replace("*", "*, 1 as new_column"), model_file)
        run_dbt()
        column_names = [column[2].lower() for column in project.run_sql(sql, fetch="all")]
        assert len(column_names) == 4
        assert "new_column" in column_names

        write_file(_MODEL_INCREMENTAL_ICEBERG_MERGE, model_file)
        run_dbt()
        column_names = [column[2].lower() for column in project.run_sql(sql, fetch="all")]
        assert len(column_names) == 3
        assert "new_column" not in column_names



import pytest
from dbt.tests.util import check_relations_equal, run_dbt
from tests.functional.adapter.statement_test.seeds import seeds_csv, statement_expected_csv

_STATEMENT_ACTUAL_SQL = """
-- {{ ref('seed') }}

{%- call statement('test_statement', fetch_result=True) -%}

  select
    count(*) as "num_records"

  from {{ ref('seed') }}

{%- endcall -%}

{% set result = load_result('test_statement') %}

{% set res_table = result['table'] %}
{% set res_matrix = result['data'] %}

{% set matrix_value = res_matrix[0][0] %}
{% set table_value = res_table[0]['num_records'] %}

select 'matrix' as source, {{ matrix_value }} as value
union all
select 'table' as source, {{ table_value }} as value
""".lstrip()


class TestStatements:
    @pytest.fixture(scope="class")
    def models(self):
        return {"statement_actual.sql": _STATEMENT_ACTUAL_SQL}

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": seeds_csv,
            "statement_expected.csv": statement_expected_csv,
        }

    def test_snowflake_statements(self, project):
        seed_results = run_dbt(["seed"])
        assert len(seed_results) == 2
        results = run_dbt()
        assert len(results) == 1

        db_with_schema = f"{project.database}.{project.test_schema}"
        check_relations_equal(
            project.adapter,
            [f"{db_with_schema}.STATEMENT_ACTUAL", f"{db_with_schema}.STATEMENT_EXPECTED"],
        )
